# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.generators.generator_context
   :platform: Unix, Windows
   :synopsis: Encpasulates contextual information used by generators.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os

import tornado



# Cache of tornado template loaders.
_TEMPLATE_LOADERS = {}

# Cache of loaded tornado templates.
_TEMPLATE_CACHE = {}


class GeneratorContext(object):
    """Encpasulates contextual information passed to generators.

    :ivar generator: Generator being executed.
    :ivar ontology: Ontology being processed.
    :ivar language: Language of generation, e.g. python.
    :ivar io_dir: Directory to which output will be written.

    """
    def __init__(self, generator, ontology, language, io_dir):
        """Instance constructor.

        """
        self.generator = generator
        self.language = language
        self.output_dir = io_dir
        self.key = generator if isinstance(generator, str) else generator.__name__
        self.ontology = ontology
        self.pkg = None
        self.cls = None
        self.enum = None
        self.code = []


    def set_package(self, pkg):
        """Sets current package being processed.

        :param pyesdoc.mp.ontologies.core.Package pkg: An ontology package being processed.

        """
        self.pkg = pkg
        self.cls = None
        self.enum = None


    def set_class(self, cls):
        """Sets current class type being processed.

        :param pyesdoc.mp.ontologies.core.Class cls: An ontology class being processed.

        """
        self.pkg = cls.package
        self.cls = cls
        self.enum = None


    def set_enum(self, enum):
        """Sets current enumerated type being processed.

        :param pyesdoc.mp.ontologies.core.Enum enum: An ontology enum being processed.

        """
        self.pkg = enum.package
        self.cls = None
        self.enum = enum


    def set_node(self, node):
        """Sets current node being processed.

        :param lxml.Element node: An XML node

        """
        # TODO: CONSIDER MOVING THIS FUNCTIONALITY TO THE pyesdoc.mp.ontologies.generators.qxml PACKAGE
        self.node = node


    def get_code(self, template, lu):
        """Generates code from a template.

        :param str template: Name of a template file.
        :module lu: Lanaugage specific utility functions passed to template.

        """
        # Set the template loader.
        template_dir = "{}/{}/templates".format(os.path.dirname(__file__), self.language)
        if template_dir not in _TEMPLATE_LOADERS:
            _TEMPLATE_LOADERS[template_dir] = tornado.template.Loader(template_dir)
        template_loader = _TEMPLATE_LOADERS[template_dir]

        # Load template & return generated code.
        template = template_loader.load(template)

        return template.generate(
            o=self.ontology,
            pkg=self.pkg,
            cls=self.cls,
            enum=self.enum,
            escape=lambda s: s.strip().replace('"', "'"),
            lu=lu
            )



# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.decoder_for_shared_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: A set of cim 1 decoders.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
from decoder_xml_utils import set_attributes
import typeset



def decode_calendar(xml, nsmap):
    """Decodes an instance of the following type: calendar.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.Calendar

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('length', False, 'int', 'child::cim:length'),
        ('range', False, decode_closed_date_range, 'child::cim:range/cim:closedDateRange'),
        ('range', False, decode_open_date_range, 'child::cim:range/cim:openDateRange'),
    ]

    return set_attributes(typeset.shared.Calendar(), xml, nsmap, decodings)


def decode_change(xml, nsmap):
    """Decodes an instance of the following type: change.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.Change

    """
    decodings = [
        ('author', False, decode_responsible_party, 'child::cim:changeAuthor'),
        ('date', False, 'datetime.datetime', 'child::cim:changeDate'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('details', True, decode_change_property, 'child::cim:detail'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('type', False, 'unicode', 'self::cim:change/@type'),
    ]

    return set_attributes(typeset.shared.Change(), xml, nsmap, decodings)


def decode_change_property(xml, nsmap):
    """Decodes an instance of the following type: change property.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.ChangeProperty

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('id', False, 'unicode', 'child::cim:id'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('value', False, 'unicode', 'child::cim:value'),
    ]

    return set_attributes(typeset.shared.ChangeProperty(), xml, nsmap, decodings)


def decode_citation(xml, nsmap):
    """Decodes an instance of the following type: citation.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.Citation

    """
    decodings = [
        ('alternative_title', False, 'unicode', 'child::gmd:alternateTitle'),
        ('alternative_title', False, 'unicode', 'child::gmd:alternateTitle/gco:CharacterString'),
        ('collective_title', False, 'unicode', 'gmd:collectiveTitle'),
        ('collective_title', False, 'unicode', 'gmd:collectiveTitle/gco:CharacterString'),
        ('date', False, 'datetime.datetime', 'child::gmd:date/gmd:CI_Date/gmd:date/gco:Date'),
        ('date_type', False, 'unicode', 'child::gmd:date/gmd:CI_Date/gmd:dateType/gmd:CI_DateTypeCode/@codeListValue'),
        ('location', False, 'unicode', 'child::gmd:otherCitationDetails'),
        ('location', False, 'unicode', 'child::gmd:otherCitationDetails/gco:CharacterString'),
        ('title', False, 'unicode', 'child::gmd:title'),
        ('title', False, 'unicode', 'child::gmd:title/gco:CharacterString'),
        ('type', False, 'unicode', 'child::gmd:presentationForm/gmd:CI_PresentationFormCode/@codeListValue'),
    ]

    return set_attributes(typeset.shared.Citation(), xml, nsmap, decodings)


def decode_closed_date_range(xml, nsmap):
    """Decodes an instance of the following type: closed date range.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.ClosedDateRange

    """
    decodings = [
        ('duration', False, 'unicode', 'child::cim:duration'),
        ('end', False, 'datetime.datetime', 'child::cim:endDate'),
        ('start', False, 'datetime.datetime', 'child::cim:startDate'),
    ]

    return set_attributes(typeset.shared.ClosedDateRange(), xml, nsmap, decodings)


def decode_compiler(xml, nsmap):
    """Decodes an instance of the following type: compiler.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.Compiler

    """
    decodings = [
        ('environment_variables', False, 'unicode', 'child::cim:compilerEnvironmentVariables'),
        ('language', False, 'unicode', 'child::cim:compilerLanguage'),
        ('name', False, 'unicode', 'child::cim:compilerName'),
        ('options', False, 'unicode', 'child::cim:compilerOptions'),
        ('type', False, 'unicode', 'child::cim:compilerType'),
        ('version', False, 'unicode', 'child::cim:compilerVersion'),
    ]

    return set_attributes(typeset.shared.Compiler(), xml, nsmap, decodings)


def decode_daily_360(xml, nsmap):
    """Decodes an instance of the following type: daily 360.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.Daily360

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('length', False, 'int', 'child::cim:length'),
        ('range', False, decode_closed_date_range, 'child::cim:range/cim:closedDateRange'),
        ('range', False, decode_open_date_range, 'child::cim:range/cim:openDateRange'),
    ]

    return set_attributes(typeset.shared.Daily360(), xml, nsmap, decodings)


def decode_data_source(xml, nsmap):
    """Decodes an instance of the following type: data source.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.DataSource

    """
    decodings = [
    ]

    return set_attributes(typeset.shared.DataSource(), xml, nsmap, decodings)


def decode_date_range(xml, nsmap):
    """Decodes an instance of the following type: date range.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.DateRange

    """
    decodings = [
        ('duration', False, 'unicode', 'child::cim:duration'),
    ]

    return set_attributes(typeset.shared.DateRange(), xml, nsmap, decodings)


def decode_doc_meta_info(xml, nsmap):
    """Decodes an instance of the following type: doc meta info.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.DocMetaInfo

    """
    decodings = [
        ('author', False, decode_responsible_party, 'child::cim:documentAuthor'),
        ('create_date', False, 'datetime.datetime', 'child::cim:documentCreationDate'),
        ('external_ids', True, 'unicode', 'child::cim:externalID'),
        ('id', False, 'unicode', 'child::cim:documentID'),
        ('update_date', False, 'datetime.datetime', 'child::cim:documentCreationDate'),
        ('version', False, 'int', 'child::cim:documentVersion'),
        ('version', False, 'int', 'self::cim:numericalExperiment/@documentVersion'),
    ]

    return set_attributes(typeset.shared.DocMetaInfo(), xml, nsmap, decodings)


def decode_doc_reference(xml, nsmap):
    """Decodes an instance of the following type: doc reference.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.DocReference

    """
    decodings = [
        ('changes', True, decode_change, 'child::cim:change'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('external_id', False, 'unicode', 'child::cim:externalID'),
        ('id', False, 'unicode', 'child::cim:id'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('type', False, 'unicode', 'child::cim:type'),
        ('version', False, 'int', 'child::cim:version'),
    ]

    return set_attributes(typeset.shared.DocReference(), xml, nsmap, decodings)


def decode_license(xml, nsmap):
    """Decodes an instance of the following type: license.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.License

    """
    decodings = [
    ]

    return set_attributes(typeset.shared.License(), xml, nsmap, decodings)


def decode_machine(xml, nsmap):
    """Decodes an instance of the following type: machine.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.Machine

    """
    decodings = [
        ('cores_per_processor', False, 'int', 'child::cim:machineCoresPerProcessor'),
        ('description', False, 'unicode', 'child::cim:machineDescription'),
        ('interconnect', False, 'unicode', 'child::cim:machineInterconnect/@value'),
        ('libraries', True, 'unicode', 'child::cim:machineLibrary'),
        ('location', False, 'unicode', 'child::cim:machineLocation'),
        ('maximum_processors', False, 'int', 'child::cim:machineMaximumProcessors'),
        ('name', False, 'unicode', 'child::cim:machineName'),
        ('operating_system', False, 'unicode', 'child::cim:machineOperatingSystem/@value'),
        ('processor_type', False, 'unicode', 'child::cim:machineProcessorType/@value'),
        ('system', False, 'unicode', 'child::cim:machineSystem'),
        ('type', False, 'unicode', '@machineType'),
        ('vendor', False, 'unicode', 'child::cim:machineVendor/@value'),
    ]

    return set_attributes(typeset.shared.Machine(), xml, nsmap, decodings)


def decode_machine_compiler_unit(xml, nsmap):
    """Decodes an instance of the following type: machine compiler unit.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.MachineCompilerUnit

    """
    decodings = [
        ('compilers', True, decode_compiler, 'child::cim:compiler'),
        ('machine', False, decode_machine, 'child::cim:machine'),
    ]

    return set_attributes(typeset.shared.MachineCompilerUnit(), xml, nsmap, decodings)


def decode_open_date_range(xml, nsmap):
    """Decodes an instance of the following type: open date range.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.OpenDateRange

    """
    decodings = [
        ('duration', False, 'unicode', 'child::cim:duration'),
        ('end', False, 'datetime.datetime', 'child::cim:endDate'),
        ('start', False, 'datetime.datetime', 'child::cim:startDate'),
    ]

    return set_attributes(typeset.shared.OpenDateRange(), xml, nsmap, decodings)


def decode_perpetual_period(xml, nsmap):
    """Decodes an instance of the following type: perpetual period.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.PerpetualPeriod

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('length', False, 'int', 'child::cim:length'),
        ('range', False, decode_closed_date_range, 'child::cim:range/cim:closedDateRange'),
        ('range', False, decode_open_date_range, 'child::cim:range/cim:openDateRange'),
    ]

    return set_attributes(typeset.shared.PerpetualPeriod(), xml, nsmap, decodings)


def decode_platform(xml, nsmap):
    """Decodes an instance of the following type: platform.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.Platform

    """
    decodings = [
        ('contacts', True, decode_responsible_party, 'child::cim:contact'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('meta', False, decode_doc_meta_info, 'self::cim:platform'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('units', True, decode_machine_compiler_unit, 'child::cim:unit'),
    ]

    return set_attributes(typeset.shared.Platform(), xml, nsmap, decodings)


def decode_property(xml, nsmap):
    """Decodes an instance of the following type: property.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.Property

    """
    decodings = [
        ('name', False, 'unicode', 'child::cim:name'),
        ('value', False, 'unicode', 'child::cim:value'),
    ]

    return set_attributes(typeset.shared.Property(), xml, nsmap, decodings)


def decode_real_calendar(xml, nsmap):
    """Decodes an instance of the following type: real calendar.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.RealCalendar

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('length', False, 'int', 'child::cim:length'),
        ('range', False, decode_closed_date_range, 'child::cim:range/cim:closedDateRange'),
        ('range', False, decode_open_date_range, 'child::cim:range/cim:openDateRange'),
    ]

    return set_attributes(typeset.shared.RealCalendar(), xml, nsmap, decodings)


def decode_relationship(xml, nsmap):
    """Decodes an instance of the following type: relationship.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.Relationship

    """
    decodings = [
    ]

    return set_attributes(typeset.shared.Relationship(), xml, nsmap, decodings)


def decode_responsible_party(xml, nsmap):
    """Decodes an instance of the following type: responsible party.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.ResponsibleParty

    """
    decodings = [
        ('abbreviation', False, 'unicode', 'child::cim:abbreviation'),
        ('address', False, 'unicode', 'child::gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:deliveryPoint'),
        ('address', False, 'unicode', 'child::gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:deliveryPoint/gco:CharacterString'),
        ('email', False, 'unicode', 'child::gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:electronicMailAddress'),
        ('email', False, 'unicode', 'child::gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:electronicMailAddress/gco:CharacterString'),
        ('individual_name', False, 'unicode', 'child::gmd:individualName'),
        ('individual_name', False, 'unicode', 'child::gmd:individualName/gco:CharacterString'),
        ('organisation_name', False, 'unicode', 'child::gmd:organisationName'),
        ('organisation_name', False, 'unicode', 'child::gmd:organisationName/gco:CharacterString'),
        ('role', False, 'unicode', 'gmd:role/gmd:CI_RoleCode/@codeListValue'),
        ('url', False, 'unicode', 'child::gmd:contactInfo/gmd:CI_Contact/gmd:onlineResource/gmd:CI_OnlineResource/gmd:linkage/gmd:URL'),
    ]

    return set_attributes(typeset.shared.ResponsibleParty(), xml, nsmap, decodings)


def decode_standard(xml, nsmap):
    """Decodes an instance of the following type: standard.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.Standard

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('name', False, 'unicode', 'child::cim:name'),
        ('version', False, 'unicode', 'child::cim:version'),
    ]

    return set_attributes(typeset.shared.Standard(), xml, nsmap, decodings)


def decode_standard_name(xml, nsmap):
    """Decodes an instance of the following type: standard name.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.shared.StandardName

    """
    decodings = [
        ('is_open', False, 'bool', '@open'),
        ('standards', True, decode_standard, 'child::cim:standard'),
        ('value', False, 'unicode', '@value'),
    ]

    return set_attributes(typeset.shared.StandardName(), xml, nsmap, decodings)





# -*- coding: utf-8 -*-

"""
.. module:: test_utils_factory.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Exposes test factory utility functions.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import random
import uuid

import arrow



def get_boolean():
    """Returns a random boolean for testing purposes.

    """
    return True


def get_date(value=None):
    """Returns a random integer for testing purposes.

    """
    if value:
        if len(value) == 4:
            return arrow.get(value, "YYYY").datetime
        else:
            return arrow.get(value).datetime
    else:
        return arrow.utcnow().datetime


def get_int(lower=0, upper=9999999, existing=None):
    """Returns a random integer for testing purposes.

    """
    result = random.randint(lower, upper)
    while existing == result:
        result = random.randint(lower, upper)
    return result


def get_float():
    """Returns a random float for testing purposes.

    """
    return random.random()


def get_string(length):
    """Returns a random string for testing purposes.

    """
    return unicode(uuid.uuid1())[:length]


def get_unicode(length=31, existing=None):
    """Returns a random unicode for testing purposes.

    """
    result = unicode(uuid.uuid1())[:length]
    while existing == result:
        result = unicode(uuid.uuid1())[:length]
    return result


def get_uuid():
    """Returns a uuid for testing purposes.

    """
    return unicode(uuid.uuid1())





# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.class_pstr.py
   :platform: Unix, Windows
   :synopsis: Represents an ontological class print string.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

class ClassPrintString(object):
    """Represents a class's print string.

    """
    def __init__(self, text, fields):
        """Instance constructor.

        :param str text: Pstr text (unformatted).
        :param str fields: Fields to be pushed into text at point of formatting.

        """
        self.text = text
        self.fields = fields




# -*- coding: utf-8 -*-

"""
.. module:: test_utils_other.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Exposes other test utility functions.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import contextlib
import inspect
import shutil
import tempfile

import pyesdoc.cv as LIB
from tests.cv.utils_assert import assert_objects



# Test term domain.
TERM_DOMAIN = "earth-system"

# Test term sub-domain.
TERM_SUBDOMAIN = "vegetation"

# Test term kind.
TERM_KIND = "tree"

# Test term kind.
TERM_NAME = "oak"

# Set of term attributes.
TERM_ATTRIBUTES = {
    'domain',
    'subdomain',
    'kind',
    'create_date',
    'status'
}


def assert_terms(term1, term2):
    """Assers equality of 2 terms.

    :param Term term1: A term to be compared against another.
    :param Term term2: A term to be compared against another.

    """
    assert_objects(term1, term2, LIB.Term)
    assert _get_object_attribute_values(term1) == _get_object_attribute_values(term1)


def create_term():
    """Creates & returns a test term.

    """
    return LIB.create(TERM_DOMAIN, TERM_SUBDOMAIN, TERM_KIND, TERM_NAME)


def get_term():
    """Returns a test term.

    """
    return LIB.get_term(TERM_DOMAIN, TERM_SUBDOMAIN, TERM_KIND, TERM_NAME)


@contextlib.contextmanager
def get_term_and_assert(assertion_callback):
    """Returns a term for testing and applies assertion callback.

    """
    term = get_term()
    try:
        yield term
    finally:
        assertion_callback(term)


def init(func, package, desc=None):
    """Initializes a test function prior to be executed.

    """
    if desc is None:
        desc = inspect.getdoc(func)
    desc = desc.strip()
    if desc[-1] == ".":
        desc = desc[:-1]
    desc = desc[0].lower() + desc[1:]
    func.description = "pyesdoc-cv-tests: {0}: {1}".format(package, desc)


def setup():
    """Performs setup functions before running a test.

    """
    LIB.init(_create_options())


def setup_and_create_termset():
    """Performs setup functions and then creates a termset prior to running a test.

    """
    setup()
    # TODO create a termset


def setup_and_create_term():
    """Performs setup functions and then creates a term prior to running a test.

    """
    setup()
    LIB.save(create_term())


def teardown():
    """Performs teardown functions after running a test.

    """
    shutil.rmtree(LIB.get_option(LIB.OPT_IO_DIR))
    LIB.cache.empty()


@contextlib.contextmanager
def get_options():
    """Returns set of options to use when running a test.

    """
    opts = _create_options()
    try:
        yield opts
    finally:
        shutil.rmtree(opts['io_dir'])


def _create_options():
    """Returns set of options to use when running a test.

    """
    return {
        'io_dir': tempfile.mkdtemp(),
        'verbose': False
    }


def _get_object_attribute_values(obj):
    """Returns collection of object attribute values.

    """
    return [getattr(obj, k) for k in dir(obj) if k not in dir(obj.__class__)]






# -*- coding: utf-8 -*-

"""
.. module:: dictionary.decoder.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Decodes a document from a python dictionary.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
from pyesdoc import ontologies
from pyesdoc.utils import convert



def _create_doc(obj, doc_type):
    """Creates & returns a document to be hydrated from a dictionary.

    """
    try:
        obj['meta']
    # ... only DocMetaInfo objects do not have a meta section.
    except KeyError:
        doc_type = ontologies.get_type_from_key('cim.2.shared.DocMetaInfo')
    else:
        if ontologies.get_type_key(doc_type) != obj['meta']['type']:
            doc_type = ontologies.get_type_from_key(obj['meta']['type'])
    finally:
        if doc_type is None:
            raise ValueError('Target decoding type is unrecognized')

    return doc_type(), ontologies.get_decoder_info(doc_type)


def _decode_simple(value, target_type, is_iterable):
    """Decodes a simple value.

    """
    if is_iterable:
        return [convert.text_to_typed_value(i, target_type) for i in value]
    return convert.text_to_typed_value(value, target_type)


def _decode(value, doc_type, is_iterable):
    """Decodes a dictionary.

    """
    def _do(obj):
        # Create doc.
        doc, doc_type_info = _create_doc(obj, doc_type)

        # Set doc attributes:
        for _name, _type, _is_iterable in doc_type_info:
            # ... skip placeholders;
            if _name not in obj:
                continue

            # ... set incoming value;
            _val = obj[_name]

            # ... nulls;
            if _val is None:
                setattr(doc, _name, [] if _is_iterable else None)

            # ... complex types;
            elif _type in ontologies.get_types():
                setattr(doc, _name, _decode(_val, _type, _is_iterable))

            # ... simple types;
            else:
                setattr(doc, _name, _decode_simple(_val, _type, _is_iterable))

        return doc

    return _do(value) if not is_iterable else [_do(i) for i in value]


def decode(as_dict):
    """Decodes a document from a dictionary.

    :param dict as_dict: A document in dictionary format.

    :returns: A pyesdoc document instance.
    :rtype: object

    """
    # Format keys.
    as_dict = convert.dict_keys(as_dict, convert.str_to_underscore_case)

    # Get document type key.
    try:
        doc_type_key = as_dict['meta']['type']
    except KeyError:
        raise KeyError('Document pyesdoc type key is invalid.')

    # Get document type.
    doc_type = ontologies.get_type_from_key(doc_type_key)
    if doc_type is None:
        raise ValueError('Document pyesdoc type key is invalid.')

    return _decode(as_dict, doc_type, False)



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.constants.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package constants.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
# Dictionary encoding.
ENCODING_DICT = "dict"

# JSON encoding.
ENCODING_JSON = "json"

# Set of supported encodings.
ENCODINGS = {
    ENCODING_DICT,
    ENCODING_JSON
}

# Governance state - the term is pending review.
GOVERNANCE_STATUS_PENDING = u'pending'

# Governance state - the term is accepted.
GOVERNANCE_STATUS_ACCEPTED = u'accepted'

# Governance state - the term is rejected.
GOVERNANCE_STATUS_REJECTED = u'rejected'

# Governance state - the term is obsolete.
GOVERNANCE_STATUS_DEPRECATED = u'obsolete'

# Set of allowed governance states.
GOVERNANCE_STATUS_SET = set([
    GOVERNANCE_STATUS_PENDING,
    GOVERNANCE_STATUS_ACCEPTED,
    GOVERNANCE_STATUS_REJECTED,
    GOVERNANCE_STATUS_DEPRECATED
])

# Regular expression for validating a canonical name.
REGEX_CANONICAL_NAME = u"/&"

# Regular expression for validating a label.
REGEX_LABEL = u"/&"

# Regular expression for validating a url.
REGEX_URL = u""



"""

A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.
"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Characteristics of the parameterised gravity waves in the atmosphere, ' \
              'whether from orography or other sources.'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': "General attributes of the gravity wave parameterisation",
    'properties': [
        ('sponge_layer', 'ENUM:sponge_layer_attributes', '1.1',
            'Sponge layer in the upper levels in order to avoid gravity wave reflection at the top.'),
        ('background', 'ENUM:background_attributes', '1.1',
            'Background wave distribution'),
        ('subgrid_scale_orography', 'ENUM:subgrid_scale_orography_attributes', '1.N',
            'Subgrid scale orography effects taken into account.'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: orographic_gravity_waves
# --------------------------------------------------------------------
DETAILS['orographic_gravity_waves'] = {
    'description': 'Gravity waves generated due to the presence of orography',
    'properties': [
        ('source_mechanisms', 'ENUM:orographic_gravity_wave_source_mechanisms', '1.N',
            'Orographic gravity wave source mechanisms'),
        ('calculation_method', 'ENUM:orographic_gravity_wave_calculation_method', '1.N',
            'Orographic gravity wave calculation method'),
        ('propagation_scheme', 'ENUM:orographic_gravity_wave_propagation_scheme', '1.1',
            'Orographic gravity wave propogation scheme'),
        ('dissipation_scheme', 'ENUM:orographic_gravity_wave_dissipation_scheme', '1.1',
            'Orographic gravity wave dissipation scheme'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: non_orographic_gravity_waves
# --------------------------------------------------------------------
DETAILS['non_orographic_gravity_waves'] = {
    'description': 'Gravity waves generated by non-orographic processes.',
    'properties': [
        ('source_mechanisms', 'ENUM:non_orographic_gravity_wave_source_mechanisms', '1.N',
            'Non-orographic gravity wave source mechanisms'),
        ('calculation_method', 'ENUM:non_orographic_gravity_wave_calculation_method', '1.N',
            'Non-orographic gravity wave calculation method'),
        ('propagation_scheme', 'ENUM:non_orographic_gravity_wave_propagation_scheme', '1.1',
            'Non-orographic gravity wave propogation scheme'),
        ('dissipation_scheme', 'ENUM:non_orographic_gravity_wave_dissipation_scheme', '1.1',
            'Non-orographic gravity wave dissipation scheme'),
        ]
    }

# --------------------------------------------------------------------
# PROCESS: ENUMERATIONS
# --------------------------------------------------------------------

ENUMERATIONS['sponge_layer_attributes'] = {
    'description': 'Gravity waves sponge layer attributes',
    'is_open': True,
    'members': [
        ('Rayleigh friction', None),
        ('Diffusive sponge layer', None)
        ]
    }

ENUMERATIONS['background_attributes'] = {
    'description': 'Gravity waves background attributes',
    'is_open': True,
    'members': [
        ('continuous spectrum', None),
        ('discrete spectrum', None),
        ]
    }

ENUMERATIONS['subgrid_scale_orography_attributes'] = {
    'description': 'Gravity waves subgrid scale orography attributes',
    'is_open': True,
    'members': [
        ('effect on drag', None),
        ('effect on lifting', None),
        ('enhanced topography', 'To enhance the generation of long waves in the atmosphere'),
        ]
    }

ENUMERATIONS['orographic_gravity_wave_source_mechanisms'] = {
    'description': 'Physical mechanisms generating orographic gravity waves.',
    'is_open': True,
    'members': [
        ('linear mountain waves', None),
        ('hydraulic jump', None),
        ('envelope orography', None),
        ('low level flow blocking', None),
        ('statistical sub-grid scale variance', None),
        ]
    }

ENUMERATIONS['orographic_gravity_wave_calculation_method'] = {
    'description': 'Calculation method for orographic gravity waves.',
    'is_open': True,
    'members': [
        ('non-linear calculation', None),
        ('more than two cardinal directions', None),
        ]
    }

ENUMERATIONS['orographic_gravity_wave_propagation_scheme'] = {
    'description': 'Type of propagation scheme for orgraphic gravity waves',
    'is_open': True,
    'members': [
        ('linear theory', None),
        ('non-linear theory', None),
        ]
    }

ENUMERATIONS['orographic_gravity_wave_dissipation_scheme'] = {
    'description': 'Type of dissipation scheme for orographic gravity waves.',
    'is_open': True,
    'members': [
        ('total wave', None),
        ('single wave', None),
        ('spectral', None),
        ('linear', None),
        ]
    }

ENUMERATIONS['non_orographic_gravity_wave_source_mechanisms'] = {
    'description': 'Physical mechanisms generating non-orographic gravity waves',
    'is_open': True,
    'members': [
        ('convection', None),
        ('precipitation', None),
        ('background spectrum', None),
        ]
    }

ENUMERATIONS['non_orographic_gravity_wave_calculation_method'] = {
    'description': 'Calculation method for non-orographic gravity waves',
    'is_open': False,
    'members': [
        ('spatially dependent', None),
        ('temporally dependent', None),
        ]
    }

ENUMERATIONS['non_orographic_gravity_wave_propagation_scheme'] = {
    'description': 'Type of propagation scheme for non-orographic gravity waves.',
    'is_open': True,
    'members': [
        ('linear theory', None),
        ('non-linear theory', None),
        ]
    }

ENUMERATIONS['non_orographic_gravity_wave_dissipation_scheme'] = {
    'description': 'Type of dissipation scheme for non-orographic gravity waves.',
    'is_open': True,
    'members': [
        ('total wave', None),
        ('single wave', None),
        ('spectral', None),
        ('linear', None),
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.factory.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates creation of domain model class instances.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
import arrow
import re
import uuid

from pyesdoc.cv import constants
from pyesdoc.cv import validation as v
from pyesdoc.cv.exceptions import ValidationError
from pyesdoc.cv.model import Term
from pyesdoc.cv.model import Authority
from pyesdoc.cv.model import Collection
from pyesdoc.cv.model import Scope



def create_authority(name, description, url, create_date=None):
	"""Instantiates, initialises & returns a term authority.

	"""
	# Validate inputs.
	v.validate_authority_name(name)
	v.validate_authority_description(description)
	v.validate_authority_url(url)

	# Format inputs.
	name = unicode(name).strip()
	description = unicode(description).strip()
	url = unicode(url).strip()

	# Instantiate.
	i = Authority()
	i.create_date = create_date or arrow.utcnow().datetime
	i.description = description
	i.label = name
	i.name = name.lower()
	i.url = url

	# Raise validation exception (if appropriate).
	if not i.is_valid:
		raise ValidationError(i.errors)

	return i


def create_scope(authority, name, description, url, create_date=None):
	"""Instantiates, initialises & returns a term scope.

	param: pyesdoc.cv.Authority authority: CV authority to which scope is bound.
	param: str name: Scope name (must be unique within authority).
	param: str description: Scope description.
	param: str url: Scope URL for further information.
	param: datetime create_date: Date upon which scope was created.

	"""
	# Validate inputs.
	v.validate(authority)
	v.validate_scope_name(name)
	v.validate_scope_description(description)
	v.validate_scope_url(url)

	# Format inputs.
	name = unicode(name).strip()
	description = unicode(description).strip()
	url = unicode(url).strip()

	# Instantiate.
	i = Scope()
	i.authority = authority
	i.create_date = create_date or arrow.utcnow().datetime
	i.description = description
	i.label = name
	i.name = name.lower()
	i.uid = uuid.uuid4()
	i.url = url

	# Append to parent & set idx.
	authority.scopes.append(i)
	i.idx = len(authority.scopes)

	# Raise validation exception (if appropriate).
	if not i.is_valid:
		raise ValidationError(i.errors)

	return i


def create_collection(scope, name, description, create_date=None):
	"""Instantiates, initialises & returns a term collection.

	param: pyesdoc.cv.Scope scope: CV scope to which collection is bound.
	param: str name: Collection name (must be unique within scope).
	param: str description: Collection description.
	param: datetime create_date: Date upon which collection was created.

	"""
	# Validate inputs.
	v.validate(scope)
	v.validate_collection_name(name)
	v.validate_collection_description(description)

	# Format inputs.
	name = unicode(name).strip()
	description = unicode(description).strip()

	# Instantiate.
	i = Collection()
	i.create_date = create_date or arrow.utcnow().datetime
	i.description = description
	i.label = name
	i.name = name.lower()
	i.scope = scope
	i.uid = uuid.uuid4()

	# Append to parent & set idx.
	scope.collections.append(i)
	i.idx = len(scope.collections)

	# Raise validation exception (if appropriate).
	if not i.is_valid:
		raise ValidationError(i.errors)

	return i


def create_term(collection, name, data=None, create_date=None):
	"""Instantiates, initialises & returns a term.

	param: pyesdoc.cv.Collection collection: The collection to which the term belongs.
	param: str name: Name of term.
	param: dict data: Arbitrary data associated with term.
	param: datetime create_date: Date upon which term was created.

	"""
	# Validate inputs.
	v.validate(collection)
	v.validate_term_name(name)
	if data is not None:
		v.validate_term_data(data)

	# Format inputs.
	name = unicode(name).strip()

	# Instantiate.
	i = Term()
	i.collection = collection
	i.create_date = create_date or arrow.utcnow().datetime
	i.label = name
	i.name = name.lower()
	i.status = constants.GOVERNANCE_STATUS_PENDING
	i.uid = uuid.uuid4()
	i.data = data

	# Append to parent & set idx.
	collection.terms.append(i)
	i.idx = len(collection.terms)

	# Raise validation exception (if appropriate).
	if not i.is_valid:
		raise ValidationError(i.errors)

	return i



# -*- coding: utf-8 -*-

"""
.. module:: xml_metafor_cim_v1.decoder.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Decodes a document from a Metafor cim v1 XML text blob.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
from lxml import etree as et


from pyesdoc._codecs.xml_metafor_cim_v1.decoder_utils import load_xml
from pyesdoc._codecs.xml_metafor_cim_v1.decoder_utils import decode_xml
from pyesdoc.ontologies import cim



# Map of decoders to document root tag names.
_decoders = {
    'cIM_Quality' : cim.v1.decoder.decode_cim_quality,
    'numericalExperiment' : cim.v1.decoder.decode_numerical_experiment,
    'CIMDocumentSet' : cim.v1.decoder.decode_document_set,
    'dataObject' : cim.v1.decoder.decode_data_object,
    'ensemble' : cim.v1.decoder.decode_ensemble,
    'gridSpec' : cim.v1.decoder.decode_grid_spec,
    'modelComponent' : cim.v1.decoder.decode_model_component,
    'numericalExperiment' : cim.v1.decoder.decode_numerical_experiment,
    'platform' : cim.v1.decoder.decode_platform,
    'simulationRun' : cim.v1.decoder.decode_simulation_run,
    'simulationComposite' : cim.v1.decoder.decode_simulation_composite,
    'statisticalModelComponent' : cim.v1.decoder.decode_statistical_model_component,
}


def decode(as_xml):
    """Decodes a pyesdoc document from passed representation.

    :param as_xml: Metafor CIM v1 XML document representation.
    :type as_xml: str | unicode | lxml.etree._ElementTree

    """
    # Defensive programming.
    xml, nsmap = load_xml(as_xml, return_nsmap=True)
    if not isinstance(xml, et._Element):
        raise TypeError('Decoding failure due to invalid Metafor CIM v1 XML.')

    # Validate decoder.
    doc_type = xml.tag.split('}')[1]
    if doc_type not in _decoders:
        raise ValueError("CIM v1 - {0} document type decoding unsupported.".format(doc_type))

    # Decode pyesdoc doc.
    return decode_xml(_decoders[doc_type], xml, nsmap, None)



"""
.. module:: factory.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Takes specialization modules and returns instances of wrapper classes.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from model import TopicPropertySpecialization
from model import TopicPropertySetSpecialization
from model import EnumSpecialization
from model import EnumChoiceSpecialization
from model import RealmSpecialization
from model import TopicSpecialization



# Map of specializations by id.
CACHE = dict()


def create_realm_specialization(spec):
    """Returns a realm specialization wrapper.

    :param spec: 4 member tuple of python modules: realm, grid, key-properties, processes.

    :returns: A realm specialization wrapper.
    :rtype: TopicSpecialization

    """
    r = _create_topic(None, spec[0], "realm", typeof=RealmSpecialization)
    r.key_properties = _create_topic(r, spec[2], "key-properties")
    r.grid = _create_topic(r, spec[1], "grid")
    r.processes = [_create_topic(r, i, "process") for i in spec[3]]

    return r


def _create_topic(parent, spec, type_key, key=None, typeof=TopicSpecialization):
    """Creates & returns a topic specialization wrapper.

    """
    topic = typeof()
    topic.type_key = type_key
    topic.spec = spec

    if isinstance(spec, dict):
        _set_topic_from_dict(parent, topic, key)
    else:
        _set_topic_from_module(parent, topic)

    topic.name_camel_case = _to_camel_case(topic.name)
    topic.name_camel_case_spaced = _to_camel_case_spaced(topic.name)
    if parent:
        topic.parent = parent
        parent.sub_topics.append(topic)

    # Cache.
    CACHE[topic.id] = topic

    return topic


def _set_topic_from_module(parent, topic):
    """Set topic specialization attributes from a module.

    """
    topic.authors = topic.spec.AUTHORS
    topic.contact = topic.spec.CONTACT
    topic.qc_status = topic.spec.QC_STATUS
    topic.description = topic.spec.DESCRIPTION
    if parent:
        topic.change_history = parent.change_history
        topic.contributors = parent.contributors
        topic.name = "_".join(topic.spec.__name__.split(".")[-1].split("_")[1:])
        topic.id = "{}.{}".format(parent.id, topic.name)
    else:
        topic.change_history = topic.spec.CHANGE_HISTORY
        topic.contributors = topic.spec.CONTRIBUTORS
        topic.name = topic.spec.__name__
        topic.id = "cmip6.{}".format(topic.name)

    # Assign properties / property sets.
    for key, obj in topic.spec.DETAILS.items():
        # ... topic toplevel properties
        if key == "toplevel":
            _set_property_collection(topic, key, obj, topic.spec.ENUMERATIONS)

        # ... topic toplevel property sets
        elif key.startswith("toplevel"):
            _set_property_set(topic, key, obj, topic.spec.ENUMERATIONS)

        # ... sub-topic properties
        elif len(key.split(":")) == 1:
            _create_topic(topic, obj, "sub-process", key)
            _set_property_collection(topic.sub_topics[-1], key, obj, topic.spec.ENUMERATIONS)

        # ... sub-topic property sets
        elif len(key.split(":")) == 2:
            for st in topic.sub_topics:
                if st.name == key.split(":")[0]:
                    _set_property_set(st, key, obj, topic.spec.ENUMERATIONS)


def _set_topic_from_dict(parent, topic, name):
    """Set topic specialization attributes from a dictionary.

    """
    topic.authors = parent.authors
    topic.contact = parent.contact
    topic.change_history = parent.change_history
    topic.contributors = parent.contributors
    topic.description = topic.spec['description']
    topic.id = "{}.{}".format(parent.id, name)
    topic.name = name
    topic.qc_status = parent.qc_status


def _set_property_set(owner, key, obj, enumerations):
    """Set attributes of a property-set attributes from a dictionary.

    """
    ps = TopicPropertySetSpecialization()
    ps.description = obj['description']
    ps.id = "{}.{}".format(owner.id, key.split(":")[-1])
    ps.key = key
    ps.name = key.split(":")[-1]
    ps.name_camel_case = _to_camel_case(ps.name)
    ps.name_camel_case_spaced = _to_camel_case_spaced(ps.name)
    ps.owner = owner
    _set_property_collection(ps, key, obj, enumerations)

    owner.property_sets.append(ps)

    # Cache.
    CACHE[ps.id] = ps


def _set_property_collection(owner, key, obj, enumerations):
    """Set a collection of topic properties from a dictionary.

    """
    for name, typeof, cardinality, description in obj['properties']:
        p = TopicPropertySpecialization()
        p.cardinality = cardinality
        p.description = description
        p.enum = _create_enum(d, typeof, enumerations) if typeof.startswith("ENUM:") else None
        p.id = "{}.{}".format(owner.id, name)
        p.key = name
        p.name = name
        p.name_camel_case = _to_camel_case(name)
        p.name_camel_case_spaced = _to_camel_case_spaced(name)
        p.owner = owner
        p.typeof = typeof

        owner.properties.append(p)

        # Cache.
        CACHE[d.id] = 


def _create_enum(detail, typeof, enumerations):
    """Creates & returns an enumeration specialzation wrapper.

    """
    key = typeof.split(":")[-1]
    obj = enumerations[key]

    e = EnumSpecialization()
    e.description = obj['description']
    e.detail = detail
    e.id = "{}.{}".format(detail.id, key)
    e.is_open = obj['is_open']
    e.label = key
    e.name = key
    e.id = key
    e.choices = [_create_enum_choice(e, i[0], i[1]) for i in obj.get('members', [])]
    if e.is_open:
        e.choices.append(_create_enum_choice(e, "Other", None))

    return e


def _create_enum_choice(enum, value, description):
    """Creates & returns an enumeration choice specialzation wrapper.

    """
    ec = EnumChoiceSpecialization()
    ec.description = description
    ec.enum = enum
    ec.id = "{}.{}".format(enum.id, value)
    ec.is_other = (value == 'Other')
    ec.value = value

    return ec


def _to_camel_case_spaced(name, separator='_'):
    """Converts passed name to camel case with space.

    :param str name: A name as specified in ontology specification.
    :param str separator: Separator to use in order to split name into constituent parts.

    """
    s = _to_camel_case(name, separator)
    r = s[0]
    for s in s[1:]:
        if s.upper() == s:
            r += " "
        r += s

    return r


def _to_camel_case(name, separator='_'):
    """Converts passed name to camel case.

    :param str name: A name as specified in ontology specification.
    :param str separator: Separator to use in order to split name into constituent parts.

    """
    r = ''
    if name is not None:
        s = name.split(separator)
        for s in s:
            if (len(s) > 0):
                r += s[0].upper()
                if (len(s) > 1):
                    r += s[1:]
    return r



# -*- coding: utf-8 -*-
"""
.. module:: content_parser.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes document content parsers.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.archive.parsing import content_parser_cmip5_metafor_q



# Set of content parsers mapped by encoding, project & source.
_PARSERS = {
    "xml-metafor-cim-v1": {
        "cmip5": {
            "metafor-q": content_parser_cmip5_metafor_q
        }
    }
}


def parse(encoding, project, source, content):
    """Parses document content according to project & source.

    :param encoding: Original content encoding.
    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).
    :param unicode content: Document content.

    """
    try:
        parser = _PARSERS[encoding][project][source]
    except KeyError:
        pass
    else:
        for func in parser.PARSERS:
            content = func(content) or content

    return content



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.exceptions.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package exceptions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
class ValidationError(ValueError):
	"""A validation error raised by the package validator.

	"""
	def __init__(self, msg):
		"""Object constructor.

		"""
		super(ValidationError, self).__init__(msg)


class InvalidPartitionError(ValueError):
	"""An error raised when user tries to access an invalid parition.

	"""
	def __init__(self, key):
		"""Object constructor.

		"""
		msg = "Unsupported pyesdoc.cv partition: {}".format(key)
		super(InvalidPartitionError, self).__init__(msg)


class InvalidOptionError(ValueError):
	"""An error raised when user tries to access an invalid option.

	"""
	def __init__(self, option):
		"""Object constructor.

		"""
		msg = "Unsupported pyesdoc.cv option: {}".format(option)
		super(InvalidOptionError, self).__init__(msg)


class InvalidAssociationError(ValueError):
	"""An error raised an invalid association is declared.

	"""
	def __init__(self, association):
		"""Object constructor.

		"""
		msg = "Unsupported association: {}".format(association)
		super(InvalidAssociationError, self).__init__(msg)



import datetime
import uuid

from pyesdoc.cv.constants import GOVERNANCE_STATUS_SET
from pyesdoc.cv.constants import REGEX_CANONICAL_NAME
from pyesdoc.cv.constants import REGEX_LABEL
from pyesdoc.cv.constants import REGEX_URL
from pyesdoc.cv.model.authority import Authority
from pyesdoc.cv.model.collection import Collection
from pyesdoc.cv.model.scope import Scope
from pyesdoc.cv.model.term import Term


# Set of supported types.
TYPES = (
    Authority,
    Collection,
    Scope,
    Term,
)

# Map of types to tuples containing validation info.
TYPE_INFO = {
    Authority: {
        ("description", unicode, "1.1"),
        ("label", unicode, "0.1", REGEX_LABEL),
        ("name", unicode, "1.1", REGEX_CANONICAL_NAME),
        ("scopes", Scope, "0.N"),
        ("url", unicode, "1.1", REGEX_URL),
    },
    Collection: {
        ("create_date", datetime.datetime, "1.1"),
        ("description", unicode, "1.1"),
        ("idx", int, "1.1"),
        ("label", unicode, "0.1", REGEX_CANONICAL_NAME),
        ("name", unicode, "1.1", REGEX_CANONICAL_NAME),
        ("scope", Scope, "1.1"),
        ("terms", Term, "0.N"),
        ("uid", uuid.UUID, "1.1"),
        ("url", unicode, "0.1", REGEX_URL),
    },
    Scope: {
        ("authority", Authority, "1.1"),
        ("collections", Collection, "0.N"),
        ("description", unicode, "0.1"),
        ("idx", int, "1.1"),
        ("label", unicode, "1.1", REGEX_LABEL),
        ("name", unicode, "1.1", REGEX_CANONICAL_NAME),
        ("uid", uuid.UUID, "1.1"),
        ("url", unicode, "1.1", REGEX_URL),
    },
    Term: {
        ("alternative_name", unicode, "0.1", REGEX_CANONICAL_NAME),
        ("alternative_url", unicode, "0.1", REGEX_URL),
        ("collection", Collection, "1.1"),
        ("create_date", datetime.datetime, "1.1"),
        ("description", unicode, "0.1"),
        ("idx", int, "1.1"),
        ("label", unicode, "1.1"),
        ("name", unicode, "1.1", REGEX_CANONICAL_NAME),
        ("parent", Term, "0.1"),
        ("status", unicode, "1.1", tuple(GOVERNANCE_STATUS_SET)),
        ("synonyms", unicode, "0.N", REGEX_CANONICAL_NAME),
        ("uid", uuid.UUID, "1.1"),
        ("url", unicode, "0.1", REGEX_URL),
    }
}



# -*- coding: utf-8 -*-

"""
.. module:: test_content_section_item.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Tests data request content section items.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import dreq



def test_type():
    """dreq-tests: config-table-attribute: verify table attribute types.

    """
    for table in dreq.config:
        for attribute in table:
            assert isinstance(attribute, dreq.ConfigTableAttribute)




# -*- coding: utf-8 -*-

"""
.. module:: test_cache.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv cache tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import pyessv

import nose

import tests.utils as tu


def _setup():
	"""Test runner setup.

	"""
	tu.setup()
	term = tu.create_term()
	pyessv.save(term)


@nose.with_setup(_setup, tu.teardown)
def test_get_term_01():
	"""pyessv-tests: cache: get term

	"""
	term = pyessv.get_term(tu.TERM_DOMAIN, tu.TERM_SUBDOMAIN, tu.TERM_KIND, tu.TERM_NAME)
	tu.assert_object(term, pyessv.Term)



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean boundary forcing'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Top level properties of boundary forcing',
    'properties': [
        ('surface_pressure', 'str', '1.1',
            'Describe how surface pressure is transmitted to ocean (via sea-ice, nothing specific,...)'),
        ('momentum_flux_correction', 'str', '0.1',
            'Describe any type of ocean surface momentum flux correction and, if applicable, how it is applied and where.'),
        ('tracers_flux_correction', 'str', '0.1',
            'Describe any type of ocean surface tracers flux correction and, if applicable, how it is applied and where.'),
        ('wave_effects', 'str', '1.1',
            'Describe if/how wave effects are modelled at ocean surface.'),
        ('river_runoff_budget', 'str', '1.1',
            'Describe how river runoff from land surface is routed to ocean and any global adjustment done.'),
        ('geothermal_heating', 'str', '1.1',
            'Describe if/how geothermal heating is present at ocean bottom.'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Momentum.
# --------------------------------------------------------------------
DETAILS['momentum'] = {
    'description': 'Key properties of momentum boundary forcing in the ocean',
    'properties': [],
    'detail_sets': [
        'bottom_friction',
        'lateral_friction'
        ]
    }

DETAILS['momentum:bottom_friction'] = {
    'description': 'Properties of momentum bottom friction in ocean',
    'properties': [
        ('type', 'ENUM:mom_bottom_friction_types', '1.1',
            'Type of momentum bottom friction in ocean'),
        ]
    }

DETAILS['momentum:lateral_friction'] = {
   'description': 'Properties of momentum lateral friction in ocean',
    'properties': [
        ('type', 'ENUM:mom_lateral_friction_types', '1.1',
            'Type of momentum lateral friction in ocean'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Tracers.
# --------------------------------------------------------------------
DETAILS['tracers'] = {
    'description': 'Key properties of tracer boundary forcing in the ocean',
    'properties': [],
    'detail_sets': [
        'sunlight_penetration',
        'fresh_water_forcing',
        ]
    }

DETAILS['tracers:sunlight_penetration'] = {
    'description': 'Properties of sunlight penetration scheme in ocean',
    'properties': [
         ('scheme', 'ENUM:sunlight_penetration_scheme_types', '1.1',
            'Type of sunlight penetration scheme in ocean'),
         ('ocean_colour', 'bool', '1.1',
            'Is the ocean sunlight penetration scheme ocean colour dependent ?'),
         ('extinction_depth', 'str', '0.1',
            'Describe and list extinctions depths for sunlight penetration scheme (if applicable).'),
        ]
    }

DETAILS['tracers:fresh_water_forcing'] = {
    'description': 'Properties of surface fresh water forcing in ocean',
    'properties': [
        ('from_atmopshere', 'ENUM:surface_fresh_water_forcing_atmos_types', '1.1',
            'Type of surface fresh water forcing from atmos in ocean'),
        ('from_sea_ice', 'ENUM:surface_fresh_water_forcing_seaice_types', '1.1',
            'Type of surface fresh water forcing from sea-ice in ocean'),
        ('forced_mode_restoring', 'str', '1.1',
            'Type of surface salinity restoring in forced mode (OMIP)'),
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['mom_bottom_friction_types'] = {
    'description': 'Type of momentum bottom friction in ocean',
    'is_open': True,
    'members': [
        ('Linear', None),
        ('Non-linear', None),
        ('Non-linear (drag function of speed of tides)', None),
        ('Constant drag coefficient', None),
        ('None', None),
        ]
    }

ENUMERATIONS['mom_lateral_friction_types'] = {
    'description': 'Type of momentum lateral friction in ocean',
    'is_open': True,
    'members': [
        ('None', None),
        ('Free-slip', None),
        ('No-slip', None),
        ]
    }

ENUMERATIONS['sunlight_penetration_scheme_types'] = {
    'description': 'Type of sunlight penetration scheme in ocean',
    'is_open': True,
    'members': [
        ('1 extinction depth', None),
        ('2 extinction depth', None),
        ('3 extinction depth', None),
        ]
    }

ENUMERATIONS['surface_fresh_water_forcing_atmos_types'] = {
    'description': 'Types of surface fresh water forcing from atmosphere in ocean',
    'is_open': True,
    'members': [
        ('Freshwater flux', None),
        ('Virtual salt flux', None),
        ]
    }

ENUMERATIONS['surface_fresh_water_forcing_seaice_types'] = {
    'description': 'Types of surface fresh water forcing from sea-ice in ocean',
    'is_open': True,
    'members': [
        ('Freshwater flux', None),
        ('Virtual salt flux', None),
        ('Real salt flux', None),
        ]
    }




import os

from pyesdoc.mp.specializations import utils_loader as loader
from pyesdoc.mp.specializations import utils_factory as factory
from pyesdoc.mp.specializations.utils_factory import CACHE


def get_model_realm_specialization(mip_era, realm):
	"""Returns model realm specializations.

	"""
	mip_era = unicode(mip_era).strip().lower()
	realm = unicode(realm).strip().lower()
	cache_key = "{}.{}".format(mip_era, realm)
	if cache_key not in CACHE:
		dpath = os.path.join(os.path.dirname(__file__), mip_era)
		specs = loader.get_realm_specializations(dpath, realm)
		factory.create_realm_specialization(specs)

	return CACHE[cache_key]


def get_property_specialization(prop_id):
	return CACHE[prop_id]




# -*- coding: utf-8 -*-
"""
.. module:: handlers_metafor_q.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes Metafor Questionnaire document processing handlers.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
# CIM v1 type keys.
_TYPE_KEY_MODEL_COMPONENT = 'cim.1.software.modelcomponent'
_TYPE_KEY_SIMULATION_RUN = 'cim.1.activity.simulationrun'
_TYPE_KEY_DOCUMENT_SET = 'cim.1.misc.documentset'

# Institute name overrides.
_INSTITUTE_OVERRIDES = {
    'CMA-BCC' : 'BCC',
    'CNRM-GAME' : 'CNRM-CERFACS',
    'GFDL' : 'NOAA-GFDL',
    'GISS' : 'NASA-GISS',
    'NASA GISS' : 'NASA-GISS',
    'NASA' : 'NASA-GISS',
    'NIMR/KMA' : 'NIMR-KMA',
    'STEVE' : 'CSIRO-QCCCE',
    'ESPS-TEAM': None
}

# Model name overrides.
_MODEL_OVERRIDES = {
    'BCC_CSM1.1' : 'BCC-CSM1.1'
}

# Default drs split.
_DRS_SPLIT_01 = '_'
_DRS_SPLIT_02 = '/'

# Name of drs tag.
_DRS_NAME_TAG = "DRS"


def _is_of_type(doc, type_key):
    """Helper function: returns document type match."""
    return doc.type_key.lower() == type_key


def _get_drs_path(doc):
    """Returns a drs path."""
    # Escape if there is no associated simulation.
    if not doc.simulation:
        return

    for external_id in doc.simulation.meta.external_ids:
        for standard in external_id.standards:
            try:
                standard.name.index(_DRS_NAME_TAG)
            except ValueError:
                continue
            else:
                return external_id.value.upper()


def _get_drs_keys(doc):
    """Returns a set of drs keys."""
    # Escape if there is no DRS defined.
    path = _get_drs_path(doc)
    if not path:
        return

    keys = []
    for key in path.split(_DRS_SPLIT_01):
        key = key.split(" ")
        keys.append(key[-1])

    return keys


def _format_model_name(name):
    """Returns parsed model name."""
    name = name.upper()
    if name in _MODEL_OVERRIDES:
        name = _MODEL_OVERRIDES[name]

    return name


def _format_institute_name(name):
    """Returns parsed institute name."""
    if name is not None:
        name = name.upper()
        if name in _INSTITUTE_OVERRIDES:
            name = _INSTITUTE_OVERRIDES[name]

    return name


def _set_drs(doc):
    """Sets DRS information."""
    if not _is_of_type(doc, _TYPE_KEY_DOCUMENT_SET):
        return

    doc.meta.drs_keys = _get_drs_keys(doc)
    if doc.meta.drs_keys:
        doc.meta.drs_path = _DRS_SPLIT_02.join(doc.meta.drs_keys)
        doc.meta.drs_keys[0] = _format_institute_name(doc.meta.drs_keys[0])
        doc.meta.drs_keys[1] = _format_model_name(doc.meta.drs_keys[1])


def _set_institute_1(doc):
    """Sets institute derived from DRS."""
    if not _is_of_type(doc, _TYPE_KEY_DOCUMENT_SET):
        return

    if doc.meta.drs_keys:
        doc.meta.institute = doc.meta.drs_keys[0]


def _set_institute_2(doc):
    """Sets institute derived from contacts."""
    if doc.meta.institute is not None:
        return

    def get_institute(contacts):
        """Returns institute code derived from contact list."""
        for contact in contacts:
            if contact.role == 'centre':
                return contact.abbreviation
        for contact in contacts:
            if contact.role == 'contact' and contact.abbreviation is not None:
                return contact.abbreviation

        return None

    try:
        doc.meta.institute = get_institute(doc.responsible_parties)
    except AttributeError:
        try:
            doc.meta.institute = get_institute(doc.contacts)
        except AttributeError:
            pass


def _set_institute_3(doc):
    """Formats institute name so as to confirm to DRS.

    """
    doc.meta.institute = _format_institute_name(doc.meta.institute)


def _propogate_meta_attributes(doc):
    """Propogates meta attributes to child documents."""
    if not _is_of_type(doc, _TYPE_KEY_DOCUMENT_SET):
        return
    if doc.meta.institute is None:
        return

    child_docs = [
        ctx.doc.experiment,
        ctx.doc.model,
        ctx.doc.platform,
        ctx.doc.simulation
        ]
    # child_docs += ctx.doc.data
    child_docs += ctx.doc.ensembles
    child_docs += ctx.doc.grids

    for child in child_docs:
        child.meta.institute = ctx.doc.meta.institute


def _set_model_name(doc):
    """Formats model name so as to confirm to DRS."""
    if not _is_of_type(doc, _TYPE_KEY_MODEL_COMPONENT):
        return

    doc.short_name = _format_model_name(doc.short_name)


def _set_model_name_inpe_hadgem2es(doc):
    """Overrides model name for the INPE HADGEM2-ES model run.

    """
    if not _is_of_type(doc, _TYPE_KEY_MODEL_COMPONENT):
        return

    if doc.meta.institute == 'INPE' and \
       doc.short_name == 'HADGEM2-ES':
        doc.short_name = 'INPE-HADGEM2-ES'


def _delete_model_properties(doc):
    """Deletes top level model properties as these were supposed to be hidden for CMIP5."""
    if _is_of_type(doc, _TYPE_KEY_MODEL_COMPONENT):
        doc.properties = []


# Set of document parsers.
PARSERS = (
    _set_drs,
    _set_institute_1,
    _set_institute_2,
    _set_institute_3,
    # _propogate_meta_attributes,
    _set_model_name,
    _set_model_name_inpe_hadgem2es,
    _delete_model_properties
    )



# -*- coding: utf-8 -*-

"""
.. module:: cim.v2.typeset_for_time_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v2.time package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid




class Calendar(object):
    """A concrete class within the cim v2 type system.

    Describes the calendar required/used in an experiment/simulation.
    This class is based on the calendar attributes and properties found in the
    CF netCDF conventions.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Calendar, self).__init__()

        self.description = None                           # unicode (0.1)
        self.month_lengths = []                           # int (0.N)
        self.name = None                                  # unicode (0.1)
        self.standard_name = None                         # time.CalendarTypes (1.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}".format(self.standard_name)


class DateTime(object):
    """A concrete class within the cim v2 type system.

    A date or time. Either in simulation time with the simulation
    calendar, or with reference to a simulation start, in which
    case the datetime is an interval after the start date.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DateTime, self).__init__()

        self.is_offset = None                             # bool (0.1)
        self.value = None                                 # unicode (1.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}(offset {})".format(self.value, self.offset)


class DatetimeSet(object):
    """An abstract class within the cim v2 type system.

    A set of times. This is an abstract class which is specialised into
    a periodic or aperiodic (irregular) list.  Note that we assume either a
    periodic set of dates which can be date and/or time, or an irregular set
    which can only be dates.

    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Instance constructor.

        """
        super(DatetimeSet, self).__init__()

        self.length = None                                # int (1.1)


class TimePeriod(object):
    """A concrete class within the cim v2 type system.

    A time period/interval (aka temporal extent).

    """
    def __init__(self):
        """Instance constructor.

        """
        super(TimePeriod, self).__init__()

        self.calendar = None                              # time.Calendar (0.1)
        self.date = None                                  # time.DateTime (0.1)
        self.date_type = None                             # time.PeriodDateTypes (1.1)
        self.length = None                                # unicode (1.1)
        self.units = None                                 # time.TimeUnits (1.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{} {}".format(self.length, self.units)


class IrregularDateset(DatetimeSet):
    """A concrete class within the cim v2 type system.

    A set of irregularly spaced times, provided as a comma separated string of yyyy-mm-dd in
     the appropriate calendar.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(IrregularDateset, self).__init__()

        self.date_set = None                              # unicode (1.1)


class RegularTimeset(DatetimeSet):
    """A concrete class within the cim v2 type system.

    A regularly spaced set of times.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(RegularTimeset, self).__init__()

        self.increment = None                             # time.TimePeriod (1.1)
        self.length = None                                # int (1.1)
        self.start_date = None                            # time.DateTime (1.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{} times from {} at {} intervals".format(self.length, self.start_date, self.increment)


class CalendarTypes(object):
    """An enumeration within the cim v2 type system.

    CF calendar types as defined in CF 1.6.
    """
    is_open = False
    members = [
        "gregorian",
        "standard",
        "proleptic_gregorian",
        "noleap",
        "365_day",
        "all_leap",
        "366_day",
        "360_day",
        "julian",
        "none"
        ]
    descriptions = [
        "Mixed Gregorian/Julian calendar as defined by Udunits",
        "Synonym for gregorian: Mixed Gregorian/Julian calendar as defined by Udunits",
        "A Gregorian calendar extended to dates before 1582-10-15. That is, a year is a leap year if either (i) it is divisible by 4 but not by 100 or (ii) it is divisible by 400.",
        "Gregorian calendar without leap years, i.e., all years are 365 days long.",
        "Synonym for noleap:Gregorian calendar without leap years, i.e., all years are 365 days long.",
        "Gregorian calendar with every year being a leap year, i.e., all years are 366 days long.",
        "Synonym for all_leap:Gregorian calendar with every year being a leap year, i.e., all years are 366 days long.",
        "All years are 360 days divided into 30 day months.",
        "Julian Calendar",
        "Perpetual time axis"
        ]


class PeriodDateTypes(object):
    """An enumeration within the cim v2 type system.

    A period date type vocabulary (used by time_period).
    """
    is_open = False
    members = [
        "unused",
        "date is start",
        "date is end"
        ]
    descriptions = [
        "Date is not used",
        "The date defines the start of the period",
        "The date is the end of the period"
        ]


class TimeUnits(object):
    """An enumeration within the cim v2 type system.

    Appropriate Time units for experiment durations in NWP and Climate Modelling.
    """
    is_open = False
    members = [
        "years",
        "months",
        "days",
        "seconds"
        ]
    descriptions = [
        "Years in current calendar",
        "Months in the current calendar",
        "86400 seconds",
        "Standard metric seconds"
        ]





# -*- coding: utf-8 -*-

"""
.. module:: html.decoder.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Decodes a document from an HTML text blob.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""



def decode(as_html):
    """Decodes a document from a html text blob.

    :param as_html: Document html representation.
    :type as_html: unicode | str

    :returns: A pyesdoc document instance.
    :rtype: object

    """
    raise NotImplementedError()



"""
.. module:: pyesdoc.mp.ontologies.core.__init__.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: pyesdoc.mp.ontologies.core package initializer.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.mp.ontologies.core.class_ import Class
from pyesdoc.mp.ontologies.core.class_constraint import ClassConstraint
from pyesdoc.mp.ontologies.core.class_pstr import ClassPrintString
from pyesdoc.mp.ontologies.core.computed_property import ComputedProperty
from pyesdoc.mp.ontologies.core.decoding import Decoding
from pyesdoc.mp.ontologies.core.enum import Enum
from pyesdoc.mp.ontologies.core.enum_member import EnumMember
from pyesdoc.mp.ontologies.core.ontology import Ontology
from pyesdoc.mp.ontologies.core.package import Package
from pyesdoc.mp.ontologies.core.property import Property
from pyesdoc.mp.ontologies.core.type_ import Type
from pyesdoc.mp.ontologies.core.factory import create_ontology



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.generators.qconfig.utils.py
   :platform: Unix, Windows
   :synopsis: helps generates qconfig file for CIM Ontology

.. moduleauthor:: Allyn Treshansky <allyn.treshansky@colorado.edu>

"""

# from esdoc_mp.ontologies.generators import generator_utils as gu
import os

QCONFIG_ATOMIC_PROPERTY_TYPE = "ATOMIC"
QCONFIG_ENUMERATION_PROPERTY_TYPE = "ENUMERATION"
QCONFIG_RELATIONSHIP_PROPERTY_TYPE = "RELATIONSHIP"

QCONFIG_ATOMIC_TYPE_MAP = {
    # TOOD: WHAT ABOUT 'email' AND 'time' ?
    "bool": "BOOLEAN",
    "date": "DATE",
    "datetime": "DATETIME",
    "float": "DECIMAL",
    "int": "INTEGER",
    "str": "STRING",
    "unicode": "STRING",
    "uri": "URL",
    "uuid": "STRING",
}


def get_property_type(prop):
    prop_type = prop.type
    # TODO: ARE THERE FINER DISTINCTIONS HERE?
    # TODO: WHAT ABOUT "is_simple" AND "is_complex"?
    if prop_type.is_class:
        return QCONFIG_RELATIONSHIP_PROPERTY_TYPE
    elif prop_type.is_enum:
        return QCONFIG_ENUMERATION_PROPERTY_TYPE
    else:
        return QCONFIG_ATOMIC_PROPERTY_TYPE


def get_atomic_type(prop):
    assert get_property_type(prop) == QCONFIG_ATOMIC_PROPERTY_TYPE
    prop_type_name = prop.type.name_of_type
    try:
        return QCONFIG_ATOMIC_TYPE_MAP[prop_type_name]
    except KeyError:
        raise ValueError("Unable to locate '{0}' in ATOMIC TYPE MAP for {1}".format(prop_type_name, prop))


# def get_enumeration_target(prop):
#     assert get_property_type(prop) == QCONFIG_ATOMIC_PROPERTY_TYPE
#     prop_type = prop.type
#     return "{0}.{1}".format(
#         prop_type.name_of_package,
#         prop_type.name_of_type
#     )


def get_relationship_target_classes(prop):
    """
    Returns a list of all classes a property can point to
    (It is a list instead of a single instance b/c of the possibility of pointing to classes which have children)
    :param prop:
    :return:
    """
    assert get_property_type(prop) == QCONFIG_RELATIONSHIP_PROPERTY_TYPE
    relationship_target_classes = []
    target_class = prop.type.cls
    if not target_class.is_abstract:
        relationship_target_classes.append(target_class)
    relationship_target_classes += [
        # TODO: DOES target_class.sub_classes RETURN THE FULL HIERARCHY OF CHILDREN?
        # TODO: OR DO I NEED TO CHECK IF target_class IS IN bases OF ANY CONCRETE CLASS IN THE HIERARCHY?
        target_class for target_class in target_class.sub_classes
        if not target_class.is_abstract
    ]
    return relationship_target_classes


def get_ontology_name(name):
    """Converts name to an ontology name suitable for qconfig.

    Keyword Arguments:
    name - name being converted.

    """
    if isinstance(name, str) == False:
        name = name.name

    return name.lower()


def get_ontology_version(version):
    """Converts version identifier to a version suitable for qconfig.

    Keyword Arguments:
    name - name of version identifier being converted.

    """
    if isinstance(version, str) == False:
        version = version.version

    return version


def _get_ontology_directory(ctx, include_version=True):
    """
    Returns ontology directory into which code is generated.

    :param GeneratorContext ctx: Generation context information.
    :param bool include_version: should version info be included in the directory

    """
    _dir = ''
    if ctx.output_dir is not None:
        _dir += ctx.output_dir + '/'
    _dir += get_ontology_name(ctx.ontology)
    if include_version:
        _dir += '/v' + get_ontology_version(ctx.ontology)
    return _dir


def _get_ontology_filename(ctx, include_version=True):
    """
    Returns ontology filename for generated code file

    :param GeneratorContext ctx: Generation context information.
    :param bool include_version: should version info be included in the filename
    :return:
    """
    ontology = ctx.ontology
    ontology_filename = get_ontology_name(ontology)
    if include_version:
        ontology_filename += "_{0}".format(get_ontology_version(ontology).replace('.', '_'))
    ontology_filename += ".json"
    return ontology_filename


def get_ontology_path(ctx):
    """
    Returns full ontology path for generated QXML file

    :param GeneratorContext ctx: Generation context information.
    :return:
    """
    ontology_path = os.path.join(
        _get_ontology_directory(ctx, include_version=True),
        _get_ontology_filename(ctx, include_version=True)
    )
    return ontology_path



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.io.write.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Writes an authority's vocabularies to file sytem.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import json
import os

from pyesdoc.cv import validation
from pyesdoc.cv.model import Authority
from pyesdoc.cv.codecs import to_json


# Manifest file name.
_MANIFEST = "MANIFEST"


def write_authority(dpath, authority):
    """Writes authority CV data to file system.

    :param str dpath: Path to directory to which the CV will be written.
    :param pyesdoc.cv.Authority authority: Authority class instance to be written to file-system.

    """
    # Validate inputs.
    if not os.path.isdir(dpath):
        raise OSError("Invalid directory.")
    if not isinstance(authority, Authority):
        raise ValueError("Invalid authority: unknown type")
    if not validation.is_valid(authority):
        raise ValueError("Invalid authority: has validation errors")

    # Write authority manifest.
    with open(os.path.join(dpath, _MANIFEST), "w") as fstream:
        fstream.write(to_json(authority))

    # Write terms.
    for scope in authority:
        for collection in scope:
            for term in collection:
                _write_term(dpath, term)


def _write_term(dpath, term):
    """Writes a term to the file system.

    """
    # Set directory.
    dpath = os.path.join(dpath, term.scope.name)
    dpath = os.path.join(dpath, term.collection.name)
    try:
        os.makedirs(dpath)
    except OSError:
        pass

    # Set file path.
    fpath = os.path.join(dpath, term.name)

    # Write term JSON file.
    with open(fpath, "w") as fstream:
        fstream.write(to_json(term))



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc._serialization.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encpasulates transformations of documents from one format to another.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from lxml.etree import _ElementTree as et
from lxml.etree import _Element as ete
import xml.etree.ElementTree as ET

from pyesdoc import constants
from pyesdoc import ontologies
from pyesdoc import _codecs



# Codecs mapped by encoding.
_CODECS = {
	constants.ENCODING_DICT: _codecs.dictionary,
	constants.ENCODING_HTML: _codecs.html,
	constants.ENCODING_JSON: _codecs.json,
	constants.ENCODING_XML: _codecs.xml,
	constants.METAFOR_CIM_XML_ENCODING: _codecs.xml_metafor_cim_v1
}

# Map of formats to allowed input types when decoding.
_DECODE_TYPE_WHITELIST = {
    constants.ENCODING_DICT : (dict, ),
    constants.ENCODING_JSON : (str, unicode),
    constants.ENCODING_XML : (str, unicode, ET.Element),
    constants.METAFOR_CIM_XML_ENCODING : (str, unicode, ET.Element, et, ete),
}


def decode(target, encoding=constants.ENCODING_JSON):
    """Decodes a document from a text blob.

    :param target: A document representation (e.g. json).
    :type target: utf-8 encoded str | unicode
    :param str encoding: A document encoding (dict|json|xml|metafor-cim-1-xml).

    :returns: A pyesdoc document instance.
    :rtype: object

    """
    _assert_decode(target, encoding)

    return _CODECS[encoding].decode(target)


def encode(target, encoding=constants.ENCODING_JSON):
    """Encodes a document.

    :param object target: pyesdoc document instance.
    :param str encoding: A document encoding.

    :returns: A formatted representation of a pyesdoc document.
    :rtype: object

    """
    try:
        iter(target)
    except TypeError:
        pass
    else:
        return [i for i in [encode(d, encoding) for d in target] if i]

    _assert_encode(target, encoding)
    encoded = _CODECS[encoding].encode(target)
    try:
        encoded = encoded.strip()
    except AttributeError:
        pass

    return encoded


def convert(doc, encoding_from, encoding_to):
    """Converts a document.

    :param unicode doc: A document representation (e.g. json).
    :param str encoding_from: A document encoding (dict|json|xml).
    :param str encoding_to: A document encoding (dict|json|xml).

    :returns: A pyesdoc document representation.
    :rtype: unicode | dict

    """
    return encode(decode(doc, encoding_from), encoding_to)


def _assert_decode(target, encoding):
    """Asserts decode input parameters.

    """
    if target is None:
        raise ValueError("Documents cannot be decoded from null objects.")

    if not encoding in _CODECS:
        raise NotImplementedError('Document decoding is unsupported :: encoding = {0}.'.format(encoding))

    if type(target) not in _DECODE_TYPE_WHITELIST[encoding]:
        err = "Document representation type ({0}) is unsupported, it must be one of {1}."
        err = err.format(type(target), _DECODE_TYPE_WHITELIST[encoding])
        raise TypeError(err)


def _assert_encode(target, encoding):
    """Asserts encode input parameters.

    """
    if target is None:
        raise ValueError("Cannot encode a null document.")

    if type(target) not in ontologies.get_types():
        raise TypeError("Unsupported document type: {0}.".format(type(target)))

    if not encoding in _CODECS:
        raise NotImplementedError('Unsupported document encoding :: {0}.'.format(encoding))



# -*- coding: utf-8 -*-

"""
.. module:: test_archive.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyesdoc document archive tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import pyesdoc
import test_utils as tu


def _test_archive_config():
    """Test archive configuration."""
    import pyesdoc.archive.config as cfg

    tu.assert_path(cfg.get_directory())
    tu.assert_int(cfg.get_projects(), 6)
    tu.assert_int(cfg.get_sources(), 4)
    tu.assert_int(cfg.get_projects()[0].feeds, 2)
    tu.assert_int(cfg.get_projects()[1].feeds, 1)
    tu.assert_int(cfg.get_projects()[2].feeds, 1)
    tu.assert_int(cfg.get_projects()[3].feeds, 1)
    tu.assert_int(cfg.get_projects()[4].feeds, 1)
    tu.assert_int(cfg.get_project_sources(), 7)


def _test_archive_doc_dirs():
    """Test archive document directories."""
    folders = list(pyesdoc.archive.yield_folders())
    for folder in folders:
        assert folder.exists, True


def _test_archive_get_counts():
    """Test archive document counts."""
    tu.assert_int(pyesdoc.archive.get_count(), 37019, assert_type=tu.COMPARE_LTE)


def test():
    """Runs set of archive unit tests."""
    for func in (
        _test_archive_config,
        _test_archive_doc_dirs,
        _test_archive_get_counts
        ):
        tu.init(func, func.__doc__[5:])
        yield func


# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.generators.python.serialization_generator.py
   :platform: Unix, Windows
   :synopsis: Generates code to support serialization.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from operator import add

from pyesdoc.mp.ontologies.generators import generator_utils as gu
from pyesdoc.mp.ontologies.generators.generator import Generator
from pyesdoc.mp.ontologies.generators.python import utils as pgu



# Generator language.
_LANG = 'python'

# Template for main module.
_TEMPLATE_MAIN = "decoder.txt"

# Template for a decoder module.
_TEMPLATE_DECODER_MODULE = 'decoder_module.txt'

# Template for a decoder function.
_TEMPLATE_DECODER_FUNCTION = "decoder_function.txt"

# Template for a decoding xml utilities.
_TEMPLATE_DECODER_XML_UTILS = "decoder_xml_utils.txt"

# Loaded templates.
_TEMPLATES = gu.load_templates(_LANG, (
    _TEMPLATE_MAIN,
    _TEMPLATE_DECODER_MODULE,
    _TEMPLATE_DECODER_FUNCTION,
    _TEMPLATE_DECODER_XML_UTILS,
))



class DecoderGenerator(Generator):
    """Generates code to support decoding.

    """
    def is_required(self, ctx):
        """Predicate determing whether code generation is required.

        :param GeneratorContext ctx: Generation context information.

        """
        return bool(ctx.ontology.decodings)


    def on_ontology_parse(self, ctx):
        """Event handler for the ontology parse event.

        :param GeneratorContext ctx: Generation context information.

        """
        ctx.code.append(
            (
                _emit_module_init(ctx.ontology),
                pgu.get_ontology_directory(ctx),
                pgu.get_module_file_name('decoder')
            )
        )
        ctx.code.append(
            (
                _TEMPLATES[_TEMPLATE_DECODER_XML_UTILS],
                pgu.get_ontology_directory(ctx),
                pgu.get_module_file_name('decoder_xml_utils')
            )
        )


    def on_package_parse(self, ctx):
        """Event handler for the package parse event.

        :param GeneratorContext ctx: Generation context information.

        """
        ctx.code.append(
            (
                _emit_module_decoder_for_pkg(ctx.ontology, ctx.pkg),
                pgu.get_ontology_directory(ctx),
                pgu.get_package_module_file_name(ctx.pkg, 'decoder')
            )
        )


def _emit_module_decoder_for_pkg(o, p):
    """Emits package decoder module."""
    code = _TEMPLATES[_TEMPLATE_DECODER_MODULE]
    code = code.replace('{module-imports}', _emit_snippet_decoder_imports(o, p))
    code = code.replace('{decoding-functions}', _emit_snippet_decoding_fns(p))

    return code


def _emit_snippet_decoder_imports(o, p):
    """Emits set of package decoder imports."""
    imports = []

    def append_import(imp):
        if imp not in imports:
            imports.append(imp)

    # Set type decoding imports.
    for type_ in [t for t in p.external_types if t.is_class]:
        imp = 'from {0} import *'.format(
            pgu.get_package_module_name(type_.name_of_package, 'decoder'))
        append_import(imp)

    if len(imports) > 0:
        return reduce(add, map(lambda i : i + gu.emit_line_return(), sorted(imports)))
    else:
        return ''


def _emit_snippet_decoding_fns(p):
    """Emits set of package class decodings."""
    code = ''
    for c in sorted(p.classes, key=lambda c: c.op_func_name):
        fn = _TEMPLATES[_TEMPLATE_DECODER_FUNCTION]
        fn = fn.replace('{class-name}', c.op_name)
        fn = fn.replace('{class-function-name}', c.op_func_name)
        fn = fn.replace('{package-name}', c.package.op_name)
        fn = fn.replace('{class-doc-name}', c.op_doc_string_name)
        fn = fn.replace('{class-decodings}', _emit_snippet_decodings(c))
        fn += gu.emit_line_return(3)
        code += fn

    return code


def _emit_snippet_decodings(c):
    """Emits a set of class decodings."""
    code = []
    for p in sorted(c.all_properties, key= lambda p: p.name):
        for dc in c.get_property_decodings(p):
            if dc.decoding is not None:
                code.append(_emit_snippet_decoding(p, dc.decoding, dc.type))

    return gu.emit(code)


def _get_decoding_function(prp, type_):
    """Returns decoding function name."""
    # ... simple/enum types - return type functional name
    #     (is directly mapped to a convertor function).
    if prp.type.is_simple or prp.type.is_enum:
        return '\'{0}\''.format(pgu.get_type_functional_name(prp.type))
    # ... complex classes - return class functional name.
    elif prp.type.is_class:
        type_name = prp.type.name if type_ is None else type_
        return _get_decoder_function_name(type_name)


def _emit_snippet_decoding(prp, decoding, type_):
    """Emits a class property decoding."""
    tmpl = '{0}(\'{1}\', {2}, {3}, \'{4}\'),'
    return tmpl.format(
        gu.emit_line_return() + gu.emit_indent(2),
        prp.name,
        prp.is_collection,
        _get_decoding_function(prp, type_),
        '' if decoding is None else decoding)


def _emit_module_init(o):
    """Emits package initializer."""
    def get_imports():
        def get_code(c):
            return "from {0} import {1}".format(
                pgu.get_package_module_name(c.package, 'decoder'),
                _get_decoder_function_name(c)) + gu.emit_line_return()

        return gu.emit(o.entities, get_code)

    code = _TEMPLATES[_TEMPLATE_MAIN]
    code = code.replace('{module-imports}', get_imports())

    return code


def _get_decoder_function_name(name):
    """Converts class name to a decoder function name.

    """
    return 'decode_{0}'.format(pgu.get_type_func_name(name))



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.io.delete.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates deleting terms from file sytem.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os

from pyesdoc.cv.model import Authority
from pyesdoc.cv.model import Term



def delete_authority(authority):
    """Deletes authority CV data from file system.

    """
    if not isinstance(authority, Authority):
        raise ValueError("Invalid authority: unknown type")

    if authority.io_path:
        os.remove(authority.io_path)


def delete_term(term):
    """Deletes term CV data from file system.

    """
    if not isinstance(term, Term):
        raise ValueError("Invalid term: unknown type")

    if term.io_path:
        os.remove(term.io_path)




# -*- coding: utf-8 -*-

"""
.. module:: parser.py
   :platform: Unix, Windows
   :synopsis: An event style specializations parser.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
class RealmSpecializationParser(object):
    """An event driven CMIP6 realm specializations parser.

    """
    def __init__(self, realm):
        """Instance constructor.

        """
        self.realm = realm


    def run(self):
        """Runs the parser raising events as it does so.

        """
        self.on_realm_parse(self.realm)
        self._parse_topic(self.realm, False)
        if self.realm.grid:
            self.on_grid_parse(self.realm.grid)
            self._parse_topic(self.realm.grid)
            self.on_grid_parsed(self.realm.grid)
        if self.realm.key_properties:
            self.on_keyproperties_parse(self.realm.key_properties)
            self._parse_topic(self.realm.key_properties)
            self.on_keyproperties_parsed(self.realm.key_properties)
        for p in self.realm.processes:
            self.on_process_parse(p)
            self._parse_topic(p)
            self.on_process_parsed(p)
        self.on_realm_parsed(self.realm)


    def _parse_topic(self, topic, parse_sub_topics=True):
        """Parses a topic.

        """
        self._parse_topic_properties(topic.properties)
        for ps in topic.property_sets:
            self.on_topic_property_set_parse(ps)
            self._parse_topic_properties(ps.properties)
            self.on_topic_property_set_parsed(ps)
        if parse_sub_topics:
            for st in topic.sub_topics:
                self.on_subprocess_parse(st)
                self._parse_topic(st)
                self.on_subprocess_parsed(st)


    def _parse_topic_properties(self, properties):
        """Parses a collection of properties associated with either a topic or a property-set.

        """
        for p in properties:
            self.on_topic_property_parse(p)
            if p.enum:
                self.on_enum_parse(p.enum)
                for ec in p.enum.choices:
                    self.on_enumchoice_parse(ec)
                    self.on_enumchoice_parsed(ec)
                self.on_enum_parsed(p.enum)
            self.on_topic_property_parsed(p)


    def on_realm_parse(self, realm):
        """On realm parse event handler.

        """
        pass


    def on_realm_parsed(self, realm):
        """On realm parsed event handler.

        """
        pass


    def on_grid_parse(self, grid):
        """On grid parse event handler.

        """
        pass


    def on_grid_parsed(self, grid):
        """On grid parsed event handler.

        """
        pass


    def on_keyproperties_parse(self, key_properties):
        """On key-properties parse event handler.

        """
        pass


    def on_keyproperties_parsed(self, key_properties):
        """On key-properties parsed event handler.

        """
        pass


    def on_process_parse(self, process):
        """On process parse event handler.

        """
        pass


    def on_process_parsed(self, process):
        """On process parsed event handler.

        """
        pass


    def on_subprocess_parse(self, subprocess):
        """On sub-process parse event handler.

        """
        pass


    def on_subprocess_parsed(self, subprocess):
        """On sub-process parsed event handler.

        """
        pass


    def on_topic_property_set_parse(self, prop_set):
        """On topic property set parse event handler.

        """
        pass


    def on_topic_property_set_parsed(self, prop_set):
        """On topic property set parsed event handler.

        """
        pass


    def on_topic_property_parse(self, prop):
        """On topic property parse event handler.

        """
        pass


    def on_topic_property_parsed(self, prop):
        """On topic property parsed event handler.

        """
        pass


    def on_enum_parse(self, enum):
        """On enum parse event handler.

        """
        pass


    def on_enum_parsed(self, enum):
        """On enum parsed event handler.

        """
        pass


    def on_enumchoice_parse(self, choice):
        """On enum choice parse event handler.

        """
        pass


    def on_enumchoice_parsed(self, choice):
        """On enum choice parsed event handler.

        """
        pass



from pyesdoc.cv.archive.accessor import load_authority
from pyesdoc.cv.archive.accessor import load_collection
from pyesdoc.cv.archive.accessor import load_scope
from pyesdoc.cv.archive.accessor import load_term



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.model.authority.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary authority, e.g. WGCM.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.cv.model.entity import Entity



class Authority(Entity):
    """An authority assuming responsibity for governance of vocabularies.

    """
    def __init__(self):
        """Instance constructor.

        """
        self.create_date = None
        self.description = None
        self.label = None
        self.name = None
        self.scopes = list()
        self.url = None


    def __repr__(self):
        """Instance representation.

        """
        return self.namespace


    def __len__(self):
        """Returns number of items in managed collection.

        """
        return len(self.scopes)


    def __iter__(self):
        """Instance iterator initializer.

        """
        return Entity.getiter(self.scopes)


    def __getitem__(self, key):
        """Returns a child section item.

        """
        return Entity.getitem(self.scopes, key)


    def __contains__(self, key):
        """Instance membership predicate.

        """
        return self[key] is not None


    @property
    def namespace(self):
        """Returns namespace used in I/O scenarios.

        """
        return self.name



# -*- coding: utf-8 -*-

"""
.. module:: xml_metafor_cim_v1.__init__.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Metafor CIM v1 XML codec.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
from pyesdoc._codecs.xml_metafor_cim_v1.decoder import decode
from pyesdoc._codecs.xml_metafor_cim_v1.encoder import encode



# -*- coding: utf-8 -*-

"""
.. module:: data_enums.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 data package enum definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def data_hierarchy_type():
    """Enumerates the level in the data hierarchy (constructed by the self-referential parent/child aggregations) is this DataObject.

    """
    return {
        'type' : 'enum',
        'is_open' : True,
        'members' : [],
    }


def data_status_type():
    """Enumerates status of a data object.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('complete', 'This DataObject is complete.'),
            ('metadataOnly', 'This DataObject is incomplete - it is described in metadata but the actual data has not yet been linked to it.'),
            ('continuouslySupplemented', 'This DataObject\'s actual data is continuously updated.'),
        ],
    }



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.archive.accessor.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates access to CV archive.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os

from pyesdoc.cv.io_mgr import read_authority


# Cached loaded CV objects.
_CACHE = {}



def load_authority(authority):
	"""Loads a CV authority from archive.

	"""
	authority = unicode(authority).strip().lower()

	_cache_authority(authority)
	try:
		return _CACHE[authority]
	except KeyError:
		pass


def load_scope(authority, scope):
	"""Loads a CV scope from archive.

	"""
	authority = unicode(authority).strip().lower()
	scope = unicode(scope).strip().lower()

	_cache_authority(authority)
	try:
		return _CACHE[authority][scope]
	except KeyError:
		pass


def load_collection(authority, scope, collection):
	"""Loads a CV collection from archive.

	"""
	authority = unicode(authority).strip().lower()
	scope = unicode(scope).strip().lower()
	collection = unicode(collection).strip().lower()

	_cache_authority(authority)
	try:
		return _CACHE[authority][scope][collection]
	except KeyError:
		pass


def load_term(authority, scope, collection, term):
	"""Loads a CV collection from archive.

	"""
	authority = unicode(authority).strip().lower()
	scope = unicode(scope).strip().lower()
	collection = unicode(collection).strip().lower()
	term = unicode(term).strip().lower()

	_cache_authority(authority)
	try:
		return _CACHE[authority][scope][collection][term]
	except KeyError:
		pass


def _cache_authority(name):
	"""Caches an authority if necessary.

	"""
	if name not in _CACHE:
		dpath = os.path.dirname(__file__)
		dpath = os.path.join(dpath, name)
		authority = read_authority(dpath)
		if authority is not None:
			_CACHE[name] = authority



# -*- coding: utf-8 -*-

"""
.. module:: quality_classes.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 quality package class definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def evaluation():
    """Creates and returns instance of evaluation class.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('date', 'datetime', '0.1'),
            ('description', 'str', '0.1'),
            ('did_pass', 'bool', '0.1'),
            ('explanation', 'str', '0.1'),
            ('specification', 'str', '0.1'),
            ('specification_hyperlink', 'str', '0.1'),
            ('type', 'str', '0.1'),
            ('type_hyperlink', 'str', '0.1'),
            ('title', 'str', '0.1'),
        ],
        'decodings' : [
            ('date', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:specification/gmd:CI_Citation/gmd:date/gmd:CI_Date/gmd:date/gco:Date'),
            ('description', 'gmd:evaluationMethodDescription/gco:CharacterString'),
            ('did_pass', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:pass/gco:Boolean'),
            ('explanation', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:explanation/gco:CharacterString'),
            ('type', 'child::gmd:result/@xlink:title'),
            ('type_hyperlink', 'child::gmd:result/@xlink:href'),
            ('specification', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:specification/@xlink:title'),
            ('specification_hyperlink', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:specification/@xlink:href'),
            ('title', 'child::gmd:result/gmd:DQ_ConformanceResult/gmd:specification/gmd:CI_Citation/gmd:title/gco:CharacterString'),
        ]
    }


def measure():
    """Creates and returns instance of measure class.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('description', 'str', '0.1'),
            ('identification', 'str', '0.1'),
            ('name', 'str', '0.1'),
        ],
        'decodings' : [
            ('description', 'child::cim:measureDescription'),
            ('identification', 'child::cim:measureIdentification/gmd:code/gco:CharacterString'),
            ('name', 'child::cim:nameOfMeasure'),

            # Hacks due to DKRZ misimplementation.
            ('description', 'parent::cim:report/child::gmd:measureDescription/gco:CharacterString'),
            ('name', 'parent::cim:report/child::gmd:nameOfMeasure/gco:CharacterString'),
        ]
    }


def cim_quality():
    """The starting point for a quality record.  It can contain any number of issues and reports.  An issue is an open-ended description of some issue about a CIM instance.  A record is a prescribed description of some specific quantitative measure that has been applied to a CIM instance.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'is_document': True,
        'properties' : [
            ('reports', 'quality.report', '0.N'),
        ],
        'decodings' : [
            ('meta', 'self::cim:cIM_Quality'),
            ('reports', 'child::cim:report'),
        ]
    }


def report():
    """Creates and returns instance of report class.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('date', 'datetime', '0.1'),
            ('evaluation', 'quality.evaluation', '1.1'),
            ('evaluator', 'shared.responsible_party', '0.1'),
            ('measure', 'quality.measure', '1.1'),
        ],
        'decodings' : [
            ('date', 'child::gmd:dateTime/gco:DateTime'),
            ('evaluation', 'self::cim:report'),
            ('evaluator', 'child::cim:evaluator'),
            ('measure', 'self::cim:report/cim:measure'),
        ]
    }





"""A realm key-properties specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean Biogeochemistry key properties'

# --------------------------------------------------------------------
# KEY PROPERTIES: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'General key properties in ocean biogeochemistry',
    'properties': [
        ('model_name', 'str', '1.1',
            'Name of ocean biogeochemistry model code (PISCES 2.0,...)'),
        ('model_family', 'str', '1.1',
            'Type of ocean biogeochemistry model'),
        ('basic_approximations', 'str', '1.1',
            'Basic approximations made in the ocean biogeochemistry',),
        ('prognostic_variables', 'str', '1.N',
            'List of prognostic variables in the ocean biogeochemistry component'),
        ('damping', 'str', '0.1',
            'Describe any tracer damping used'),

        ]
    }

# --------------------------------------------------------------------
# KEY PROPERTIES: details
# --------------------------------------------------------------------
DETAILS['time_stepping_framework'] = {
    'description': 'Time stepping frameowrk in ocean biogeochemistry',
    'properties' : []
    }

DETAILS['time_stepping_framework:passive_tracers_transport'] = {
    'description': 'Time stepping method for passive tracers transport in ocean biogeochemistry',
    'properties' : [
        ('passive_tracers_transport_method', 'ENUM:passive_tracers_transport', '1.1',
            'Time stepping framework for passive tracers'),
        ('passive_tracers_timestep', 'int', '0.1',
            'Time step for passive tracers (if different from ocean)'),
        ]
    }

DETAILS['time_stepping_framework:biology_sources_sinks'] = {
    'description': 'Time stepping framework for biology sources and sinks in ocean biogeochemistry',
    'properties' : [
        ('biology_sources_sinks_transport_method', 'ENUM:passive_tracers_transport', '1.1',
            'Time stepping framework for biology sources and sinks'),
        ('passive_tracers_timestep', 'int', '0.1',
            'Time step for biology sources and sinks (if different from ocean)'),
        ]
    }

DETAILS['transport_scheme'] = {
    'description': 'Transport scheme in ocean biogeochemistry',
    'properties' : [
        ('type', 'ENUM:transport_types', '1.1',
            'Type of transport scheme'),
        ('scheme', 'bool', '1.1',
            'Is transport scheme same as that of ocean model ?'),
        ('different_scheme', 'str', '0.1',
            'Decribe transport scheme if differen than that of ocean model'),
        ]
    }

DETAILS['boundary_forcing'] = {
    'description': 'Properties of biogeochemistry boundary forcing',
    'properties': [
        ('atmospheric_deposition', 'ENUM:sources_atmos_deposition', '1.1',
            'Describe how atmospheric deposition is modeled'),
        ('river_input', 'ENUM:sources_river_input', '1.1',
            'Describe how river input is modeled'),
        ('sediment_boundary_conditions', 'str', '0.1',
            'List which sediments are speficied from boundary condition'),
        ('sediment_from_explicit_model', 'str', '0.1',
            'List which sediments are speficied from explicit sediment model'),
        ]
    }

DETAILS['gas_exchange'] = {
    'description': 'Properties of gas exchange in ocean biogeochemistry ',
    'properties': [
        ('co2_exchange', 'bool', '1.1',
            'Is CO2 gas exchange modeled ?'),
        ('co2_type', 'ENUM:gas_exchange_types', '0.1',
            'Describe CO2 gas exchange'),
        ('o2_exchange', 'bool', '1.1',
            'Is O2 gas exchange modeled ?'),
        ('o2_type', 'ENUM:gas_exchange_types', '0.1',
            'Describe O2 gas exchange'),
        ('dms_exchange', 'bool', '1.1',
            'Is DMS gas exchange modeled ?'),
        ('dms_type', 'str', '0.1',
            'Specify DMS gas exchange scheme type'),
        ('n2_exchange', 'bool', '1.1',
            'Is N2 gas exchange modeled ?'),
        ('n2_type', 'str', '0.1',
            'Specify N2 gas exchange scheme type'),
        ('n2o_exchange', 'bool', '1.1',
            'Is N2O gas exchange modeled ?'),
        ('n2o_type', 'str', '0.1',
            'Specify N2O gas exchange scheme type'),
        ('CO_exchange', 'bool', '1.1',
            'Is CO gas exchange modeled ?'),
        ('CO_type', 'str', '0.1',
            'Specify CO gas exchange scheme type'),
        ('other_gases', 'str', '0.1',
            'Specify any other gas exchange'),
        ]
    }

DETAILS['carbon_chemistry'] = {
    'description': 'Properties of carbon chemistry biogeochemistry',
    'properties': [
        ('type', 'ENUM:carbon_chemistry', '1.1',
            'Describe how carbon chemistry is modeled'),
        ('ph_scale', 'ENUM:ph_scale', '0.1',
            'If NOT OMIP protocol, describe pH scale.'),
        ('constants', 'str', '0.1',
            'If NOT OMIP protocol, list carbon chemistry constants.'),
        ]
    }


# --------------------------------------------------------------------
# KEY PROPERTIES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['passive_tracers_transport'] = {
    'description': 'Types of time stepping framework for passive tracers ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('use ocesn model transport time step', None),
        ('use specific time step', None),
        ]
    }

ENUMERATIONS['transport_types'] = {
    'description': 'Types of transport in ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('Offline', None),
        ('Online', None),
        ]
    }

ENUMERATIONS['sources_atmos_deposition'] = {
    'description': 'Type of atmospheric deposition in ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('from file', None),
        ('from Atmospheric Chemistry model', None),
        ]
    }

ENUMERATIONS['sources_river_input'] = {
    'description': 'Type of river input in ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('from file', None),
        ('from Land Surface model', None),
        ]
    }

ENUMERATIONS['gas_exchange_types'] = {
    'description': 'Type of gas exchange in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('OMIP protocol', None),
        ]
    }

ENUMERATIONS['carbon_chemistry'] = {
    'description': 'Type of carbon chemistry in ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('OMIP protocol', None),
        ('Other protocol', None),
        ]
    }

ENUMERATIONS['ph_scale'] = {
    'description': 'Type of carbon chemistry pH scale in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('Sea water', None),
        ('Free', None),
        ]
    }





# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.model.entity.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: An entity within the pyesdoc.cv domain model.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import uuid

import pyesdoc
from pyesdoc.cv.constants import ENCODING_DICT
from pyesdoc.cv.constants import ENCODING_JSON



class Entity(object):
    """An entity within the pyesdoc.cv domain model.

    """
    def validate(self):
        """Validates instance.

        :returns: Set of validation errrors.
        :rtype: set

        """
        # N.B. just-in-time import to avaoid circular references.
        from pyesdoc.cv import validation as v

        return v.validate(self)


    @property
    def errors(self):
        """Returns set of validation errors.

        """
        return sorted(list(self.validate()))


    @property
    def is_valid(self):
        """Gets flag indicating validity status.

        """
        return len(self.validate()) == 0


    @property
    def as_dict(self):
        """Returns dictionary representation of term.

        """
        return pyesdoc.cv.encode(self, ENCODING_DICT)


    @property
    def as_json(self):
        """Returns json representation of term.

        """
        return pyesdoc.cv.encode(self, ENCODING_JSON)


    @classmethod
    def from_json(cls, encoded):
        """Returns json representation of term.

        """
        return pyesdoc.cv.decode(encoded, ENCODING_JSON)


    @staticmethod
    def getiter(items):
        """Returns an iterator over a managed collection.

        """
        return iter(sorted(items, key=lambda i: i if isinstance(i, (str, unicode)) else i.name))


    @staticmethod
    def getitem(items, key):
        """Returns an item from a managed collections.

        """
        # Set comparator to be used.
        if isinstance(key, int):
            comparator = lambda i: i.idx
        elif isinstance(key, uuid.UUID):
            comparator = lambda i: i.uid
        else:
            key = unicode(key).strip().lower()
            try:
                uuid.UUID(key)
            except ValueError:
                comparator = lambda i: i.name
            else:
                comparator = lambda i: unicode(i.uid)

        # Return first matching item.
        for item in items:
            if comparator(item) == key:
                return item




# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.ontologies.cim.v1.decoder.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: A set of cim 1 decoders.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
from decoder_for_activity_package import decode_downscaling_simulation
from decoder_for_activity_package import decode_ensemble
from decoder_for_activity_package import decode_numerical_experiment
from decoder_for_activity_package import decode_simulation_composite
from decoder_for_activity_package import decode_simulation_run
from decoder_for_data_package import decode_data_object
from decoder_for_grids_package import decode_grid_spec
from decoder_for_misc_package import decode_document_set
from decoder_for_quality_package import decode_cim_quality
from decoder_for_shared_package import decode_platform
from decoder_for_software_package import decode_model_component
from decoder_for_software_package import decode_processor_component
from decoder_for_software_package import decode_statistical_model_component






# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.decoder_for_grids_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: A set of cim 1 decoders.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
from decoder_xml_utils import set_attributes
from decoder_for_shared_package import *
import typeset



def decode_coordinate_list(xml, nsmap):
    """Decodes an instance of the following type: coordinate list.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.grids.CoordinateList

    """
    decodings = [
        ('has_constant_offset', False, 'bool', '@hasConstantOffset'),
        ('length', False, 'int', '@length'),
        ('uom', False, 'unicode', '@uom'),
    ]

    return set_attributes(typeset.grids.CoordinateList(), xml, nsmap, decodings)


def decode_grid_extent(xml, nsmap):
    """Decodes an instance of the following type: grid extent.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.grids.GridExtent

    """
    decodings = [
        ('maximum_latitude', False, 'unicode', 'child::cim:latMax'),
        ('maximum_longitude', False, 'unicode', 'child::cim:lonMax'),
        ('minimum_latitude', False, 'unicode', 'child::cim:latMin'),
        ('minimum_longitude', False, 'unicode', 'child::cim:lonMin'),
        ('units', False, 'unicode', 'child::cim:units'),
    ]

    return set_attributes(typeset.grids.GridExtent(), xml, nsmap, decodings)


def decode_grid_mosaic(xml, nsmap):
    """Decodes an instance of the following type: grid mosaic.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.grids.GridMosaic

    """
    decodings = [
        ('citations', True, decode_citation, 'child::cim:citationList/cim:citation'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('has_congruent_tiles', False, 'bool', '@congruentTiles'),
        ('id', False, 'unicode', '@id'),
        ('is_leaf', False, 'bool', '@isLeaf'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('mnemonic', False, 'unicode', 'child::cim:mnemonic'),
        ('mosaic_count', False, 'int', '@numMosaics'),
        ('mosaics', True, decode_grid_mosaic, 'child::cim:gridMosaic'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('tile_count', False, 'int', '@numTiles'),
        ('tiles', True, decode_grid_tile, 'child::cim:gridTile'),
        ('type', False, 'unicode', '@gridType'),
    ]

    return set_attributes(typeset.grids.GridMosaic(), xml, nsmap, decodings)


def decode_grid_property(xml, nsmap):
    """Decodes an instance of the following type: grid property.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.grids.GridProperty

    """
    decodings = [
        ('name', False, 'unicode', 'child::cim:name'),
        ('value', False, 'unicode', 'child::cim:value'),
    ]

    return set_attributes(typeset.grids.GridProperty(), xml, nsmap, decodings)


def decode_grid_spec(xml, nsmap):
    """Decodes an instance of the following type: grid spec.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.grids.GridSpec

    """
    decodings = [
        ('esm_exchange_grids', True, decode_grid_mosaic, 'child::cim:esmExchangeGrid'),
        ('esm_model_grids', True, decode_grid_mosaic, 'child::cim:esmModelGrid'),
        ('meta', False, decode_doc_meta_info, 'self::cim:gridSpec'),
    ]

    return set_attributes(typeset.grids.GridSpec(), xml, nsmap, decodings)


def decode_grid_tile(xml, nsmap):
    """Decodes an instance of the following type: grid tile.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.grids.GridTile

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('discretization_type', False, 'unicode', '@discretizationType'),
        ('extent', False, decode_grid_extent, 'child::cim:extent'),
        ('geometry_type', False, 'unicode', '@geometryType'),
        ('horizontal_resolution', False, decode_grid_tile_resolution_type, 'child::cim:horizontalResolution'),
        ('id', False, 'unicode', '@id'),
        ('is_conformal', False, 'bool', '@isConformal'),
        ('is_regular', False, 'bool', '@isRegular'),
        ('is_terrain_following', False, 'bool', '@isTerrainFollowing'),
        ('is_uniform', False, 'bool', '@isUniform'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('mnemonic', False, 'unicode', 'child::cim:mnemonic'),
        ('nx', False, 'int', '@nx'),
        ('ny', False, 'int', '@ny'),
        ('nz', False, 'int', '@nz'),
        ('refinement_scheme', False, 'unicode', '@refinementScheme'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('vertical_resolution', False, decode_grid_tile_resolution_type, 'child::cim:verticalResolution'),
        ('zcoords', False, decode_vertical_coordinate_list, 'child::cim:zcoords'),
    ]

    return set_attributes(typeset.grids.GridTile(), xml, nsmap, decodings)


def decode_grid_tile_resolution_type(xml, nsmap):
    """Decodes an instance of the following type: grid tile resolution type.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.grids.GridTileResolutionType

    """
    decodings = [
        ('description', False, 'unicode', '@description'),
        ('properties', True, decode_grid_property, 'child::cim:property'),
    ]

    return set_attributes(typeset.grids.GridTileResolutionType(), xml, nsmap, decodings)


def decode_simple_grid_geometry(xml, nsmap):
    """Decodes an instance of the following type: simple grid geometry.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.grids.SimpleGridGeometry

    """
    decodings = [
    ]

    return set_attributes(typeset.grids.SimpleGridGeometry(), xml, nsmap, decodings)


def decode_vertical_coordinate_list(xml, nsmap):
    """Decodes an instance of the following type: vertical coordinate list.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.grids.VerticalCoordinateList

    """
    decodings = [
        ('form', False, 'unicode', '@coordinateForm'),
        ('has_constant_offset', False, 'bool', '@hasConstantOffset'),
        ('length', False, 'int', '@length'),
        ('properties', True, decode_grid_property, 'child::cim:property'),
        ('type', False, 'unicode', '@coordinateType'),
        ('uom', False, 'unicode', '@uom'),
    ]

    return set_attributes(typeset.grids.VerticalCoordinateList(), xml, nsmap, decodings)





# -*- coding: utf-8 -*-
"""
.. module:: pyesdoc.drq.__init__.py

   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Library constructor.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.drq.utils import initialize
from pyesdoc.drq.content.query import pluck as query
from pyesdoc.drq.content.section import Section
from pyesdoc.drq.content.section_item import SectionItem
from pyesdoc.drq.definition.table import Table
from pyesdoc.drq.definition.table_attribute import TableAttribute




# -*- coding: utf-8 -*-

"""
.. module:: cim.v2.typeset.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Typeset information for the cim v2 ontology.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
from typeset_for_activity_package import *
from typeset_for_data_package import *
from typeset_for_designing_package import *
from typeset_for_drs_package import *
from typeset_for_platform_package import *
from typeset_for_science_package import *
from typeset_for_shared_package import *
from typeset_for_software_package import *
from typeset_for_time_package import *


import typeset_for_activity_package as activity
import typeset_for_data_package as data
import typeset_for_designing_package as designing
import typeset_for_drs_package as drs
import typeset_for_platform_package as platform
import typeset_for_science_package as science
import typeset_for_shared_package as shared
import typeset_for_software_package as software
import typeset_for_time_package as time



# Module exports.
__all__ = [
    # Packages
    activity,
    data,
    designing,
    drs,
    platform,
    science,
    shared,
    software,
    time,
    # Classes
    Activity,
    AxisMember,
    Conformance,
    Ensemble,
    EnsembleAxis,
    EnsembleMember,
    ParentSimulation,
    UberEnsemble,
    Dataset,
    Downscaling,
    InputDataset,
    Simulation,
    VariableCollection,
    AxisMember,
    DomainRequirements,
    EnsembleRequirement,
    ForcingConstraint,
    InitialisationRequirement,
    MultiEnsemble,
    NumericalExperiment,
    NumericalRequirement,
    OutputRequirement,
    Project,
    SimulationPlan,
    StartDateEnsemble,
    TemporalConstraint,
    DrsAtomicDataset,
    DrsEnsembleIdentifier,
    DrsExperiment,
    DrsGeographicalIndicator,
    DrsPublicationDataset,
    DrsSimulationIdentifier,
    DrsTemporalIdentifier,
    ComponentPerformance,
    ComputePool,
    Machine,
    Partition,
    Performance,
    StoragePool,
    StorageVolume,
    Model,
    Realm,
    Topic,
    TopicProperty,
    TopicPropertySet,
    Citation,
    DocMetaInfo,
    DocReference,
    ExtraAttribute,
    OnlineResource,
    Party,
    QualityReview,
    Responsibility,
    TextBlob,
    ComponentBase,
    Composition,
    DevelopmentPath,
    EntryPoint,
    Gridspec,
    Implementation,
    SoftwareComponent,
    Variable,
    Calendar,
    DateTime,
    DatetimeSet,
    IrregularDateset,
    RegularTimeset,
    TimePeriod,
    # Enums
    ConformanceType,
    DataAssociationTypes,
    EnsembleTypes,
    ExperimentalRelationships,
    ForcingTypes,
    NumericalRequirementScope,
    DrsFrequencyTypes,
    DrsGeographicalOperators,
    DrsRealms,
    DrsTimeSuffixes,
    StorageSystems,
    VolumeUnits,
    ModelTypes,
    NilReason,
    QualityStatus,
    RoleCode,
    TextBlobEncoding,
    CouplingFramework,
    ProgrammingLanguage,
    CalendarTypes,
    PeriodDateTypes,
    TimeUnits,
]




# -*- coding: utf-8 -*-
from pyesdoc._extensions.cim.v2 import designing_numerical_experiment
from pyesdoc._extensions.cim.v2 import designing_project



# Supported extenders keyed by document type.
SUPPORTED = {
	"cim.2.designing.numericalexperiment": designing_numerical_experiment,
	"cim.2.designing.project": designing_project,
}



# -*- coding: utf-8 -*-

"""
.. module:: shared_classes.py
   :synopsis: Set of CIM v2 ontology type definitions.

"""


def text_blob():
    """Provides a text class which supports plaintext, html, and
    friends (or will do).

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'pstr': ('{}', ('content',)),
        'properties': [
            ('content', 'str', '1.1',
                "Raw content (including markup)."),
            ('encoding', 'shared.text_blob_encoding', '1.1',
                "Content encoding.")
        ]
    }


def citation():
    """An academic reference to published work.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'is_document': True,
        'properties': [
            ('abstract', 'str', '0.1',
                "Abstract providing high level reference overview."),
            ('citation_detail', 'str', '0.1',
                "Complete citation string as would appear in a bibliography."),
            ('collective_title', 'str', '0.1',
                "Citation collective title, i.e. with all authors declared."),
            ('context', 'str', '0.1',
                "Brief text description of why this resource is being cited."),
            ('doi', 'str', '0.1',
                "Digital Object Identifier, if it exists."),
            ('title', 'str', '0.1',
                "Citation short title."),
            ('type', 'str', '0.1',
                "Citation type."),
            ('url', 'shared.online_resource', '0.1',
                "Location of electronic version.")
        ]
    }


def extra_attribute():
    """An extra attribute with key and value needed to encode further information
    not in the CIM2 domain model or specialisation. Typical use case: in parsing
    data and encoding attributes found in data.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'pstr': ('{}:{}', ('key', 'value')),
        'properties': [
            ('doc', 'str', '0.1',
                "Documentation associated with this key."),
            ('key', 'str', '1.1',
                "Name of attribute."),
            ('type', 'str', '0.1',
                "If a non-string type, provide type."),
            ('value', 'str', '1.1',
                "Value associated with key.")
        ]
    }


def nil_reason():
    """Provides an enumeration of possible reasons why a property has not been defined
    Based on GML nilReason as discussed here: https://www.seegrid.csiro.au/wiki/AppSchemas/NilValues.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("nil:inapplicable", "There is no value"),
            ("nil:missing", "The correct value is not available. Furthermore, a correct value may not exist"),
            ("nil:template", "The value will be available later"),
            ("nil:unknown", "The correct value is not known at this time. However, a correct value probably exists"),
            ("nil:withheld", "The value is not divulged")
        ]
    }


def online_resource():
    """A minimal approximation of ISO19115 CI_ONLINERESOURCE, provides a link and details
    of how to use that link.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('description', 'str', '0.1',
                "Detail of how to access the resource."),
            ('linkage', 'str', '1.1',
                "A URL."),
            ('name', 'str', '1.1',
                "Name of online resource."),
            ('protocol', 'str', '0.1',
                "Protocol to use at the linkage.")
        ]
    }


def party():
    """Implements minimal material for an ISO19115-1 (2014) compliant party.
    For our purposes this is a much better animal than the previous responsibleParty 
    which munged roles together with people. Note we have collapsed CI_Contact,
    CI_Individual and CI_Organisation as well as the abstract CI_Party.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'is_document': True,
        'properties': [
            ('address', 'str', '0.1',
                "Institutional address."),
            ('email', 'str', '0.1',
                "Email address."),
            ('name', 'str', '0.1',
                "Name of person or organisation."),
            ('orcid_id', 'str', '0.1',
                "Orcid ID if available."),
            ('is_organisation', 'bool', '0.1',
                "True if an organisation not a person."),
            ('url', 'shared.online_resource', '0.1',
                "URL of person or institution.")
        ]
    }


def quality_review():
    """Assertions as to the completeness and quality of a document.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'is_document': True,
        'properties': [
            ('date', 'str', '1.1',
                "Date upon which review was made."),
            ('metadata_reviewer', 'linked_to(shared.party)', '1.1',
                "Party who made the metadata quality assessment."),
            ('quality_description', 'str', '1.1',
                "Assessment of quality of target document."),
            ('quality_status', 'shared.quality_status', '0.1',
                "Status from a controlled vocabulary."),
            ('target_document', 'shared.doc_reference', '1.1',
                "This is the document about which quality is asserted.")
        ]
    }


def quality_status():
    """Assertion as to the review status of document.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("incomplete", "Currently being worked on"),
            ("finalised", "Author has completed document, prior to review"),
            ("under_review", "Document is being reviewed"),
            ("reviewed", "Document has been formally reviewed and assessed as complete and accurate")
        ]
    }


def responsibility():
    """Implements the ISO19115-1 (2014) CI_Responsibility (which replaces
    responsibleParty). Combines a person and their role in doing something.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'pstr': ('{}:{}', ('role', 'parties')),
        'properties': [
            ('parties', 'linked_to(shared.party)', '1.N',
                "Parties delivering responsibility."),
            ('role', 'shared.role_code', '1.1',
                "Role that the party plays or played."),
            ('when', 'time.time_period', '0.1',
                "Period when role was active, if no longer.")
        ]
    }


def role_code():
    """Responsibility role codes: roles that a party may play in delivering a responsibility.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("Principal Investigator", "Key party responsible for the existence of the resource"),
            ("originator", "Original source for the resource if obtained from elsewhere"),
            ("author", "Party who created (or co-created) resource"),
            ("collaborator", "Contributor to the production of the resource"),
            ("publisher", "Party who published the resource"),
            ("owner", "Party with legal ownership of the resource"),
            ("processor", "Party who has taken part in the workflow that resulted in this resource"),
            ("distributor", "Party who distributes the resource"),
            ("sponsor", "Party who has invested in the production of the resource"),
            ("user", "Party who uses the resource"),
            ("point of contact", "Party who can be contacted for acquiring knowledge about or acquisition of the resource"),
            ("resource provider", "Party that supplies the resource"),
            ("custodian", "Party that accepts accountability and responsibility for the source resource"),
            ("metadata_reviewer", "Party who carried out an independent review of (this) documentation"),
            ("metadata_author", "Party who created (this) documentation")
        ]
    }


def text_blob_encoding():
    """Types of text understood by the CIM notebook. Currently only
    plaintext, but we expect safe HTML to be supported as soon as practicable.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("ascii", "Normal plain text")
        ]
    }


# def reference():
#     """ An external citation target which can have a context associated with it. """
#     return {
#         'type': 'class',
#         'base': None,
#         'is_abstract': False,
#         'properties': [
#             ('document', 'shared.citation_target', '1.1', 'Reference Target'),
#             ('context', 'shared.text_blob', '0.1', 'Brief text description of why this resource is being cited'),
#         ],
#     }


# def citation_target():
#     """ A real world document, could be a book, a journal article, a manual, a web page ... it might or might
#     not be online, although preferably it would be."""
#     return {
#         'type': 'class',
#         'base': None,
#         'is_abstract': False,
#         'is_document': True,
#         'pstr': ('%s', ('name', )),
#         'properties': [
#             ('title', 'str', '1.1', 'Title or name of the document'),
#             ('name', 'str', '1.1', 'A name for the citation: short hand description, e.g. Meehl et al (2014)'),
#             ('citation_detail', 'str', '0.1', 'Complete citation string as would appear in a bibliography.'),
#             ('online_at', 'shared.online_resource', '0.1', 'Location of electronic version'),
#             ('doi', 'str', '0.1', 'Digital Object Identifier, if it exists.')
#         ]
#     }



"""

A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.
"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Characteristics of the atmosphere radiation process'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': "Radiative agents in the atmosphere model",
    'properties': [
        ('aerosols', 'ENUM:aerosol_types', '1.N',
            'Aerosols whose radiative effect is taken into account in the atmosphere model'),
        ('greenhouse_gases', 'ENUM:ghg_types', '1.N',
            'Greenhouse gases whose radiative effect is taken into account in the atmosphere model'),
        ('cloud_ice', 'ENUM:cloud_ice_properties', '1.N',
            'Radiative properties of ice crystals in clouds'),
        ('cloud_liquid', 'ENUM:cloud_liquid_properties', '1.N',
            'Radiative properties of liquid droplets in clouds'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: longwave_radiation
# --------------------------------------------------------------------
DETAILS['longwave_radiation'] = {
    'description': 'Properties of the longwave radiation scheme',
    'properties': [
        ('scheme_type', 'ENUM:longwave_scheme_type', '1.1',
            'Longwave radiation scheme type'),
        ('scheme_method', 'ENUM:longwave_scheme_method', '1.1',
            'Longwave radiation scheme method'),
        ('spectral_intervals', 'int', '1.1',
            'Longwave radiation scheme spectral intervals'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: shortwave_radiation
# --------------------------------------------------------------------
DETAILS['shortwave_radiation'] = {
    'description': 'Properties of the shortwave radiation scheme',
    'properties': [
        ('scheme_type', 'ENUM:shortwave_scheme_type', '1.1',
            'Shortwave radiation scheme type'),
        ('spectral_intervals', 'int', '1.1',
            'Shortwave radiation scheme spectral intervals'),
        ]
    }


# --------------------------------------------------------------------
# SUB-PROCESS: single_scattering_properties
# --------------------------------------------------------------------
DETAILS['single_scattering_properties'] = {
    'description': 'Single scattering properties of ice clouds',
    'properties': [
        ('single_scattering', 'ENUM:single_scattering_properties_methods', '0.N',
            'Methods for calculating single scattering properties of ice crystals'),
        ]
    }

# --------------------------------------------------------------------
# PROCESS: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['aerosol_types'] = {
    'description': 'Aerosols whose radiative effect is taken into account in the atmospheric model.',
    'is_open': True,
    'members': [
        ('sulphate', None),
        ('nitrate', None),
        ('sea salt', None),
        ('dust', None),
        ('ice', None),
        ('organic', None),
        ('BC (black carbon / soot)', None),
        ('SOA (secondary organic aerosols)', None),
        ('POM (particulate organic matter)', None),
        ('polar stratospheric ice', None),
        ('NAT (nitric acid trihydrate)', None),
        ('NAD (nitric acid dihydrate)', None),
        ('STS (supercooled ternary solution aerosol particle)', None),
        ]
    }

ENUMERATIONS['ghg_types'] = {
    'description': 'Greenhouse gases whose radiative effect is taken into account in the atmospheric model',
    'is_open': True,
    'members': [
        ('H2O', None),
        ('CO2', None),
        ('CH4', None),
        ('N2O', None),
        ('O3', None),
        ('CFCs', None),
        ('HFCs', None),
        ('PFCs', None),
        ('SF6', None),
        ('NF3', None),
        ]
    }


ENUMERATIONS['cloud_ice_properties'] = {
    'description': 'Radiative properties of ice crystals in clouds',
    'is_open': True,
    'members': [
        ('treated as spherical particles', None),
        ('effective crystal radius', None),
        ('treated as non-spherical particles', None),
        ('mean projected area',
            'randomly oriented irregular ice crystals present a greater mean projected area than spheres'),
        ('bi-modal size distribution',
            'small mode diameters: a few tens of microns, large mode diameters: typically hundreds of microns'),
        ('ensemble of ice crystals', 'complex shapes represented with an ensemble of symmetric shapes'),
        ('ice water path', 'Integrated ice water path through the cloud kg m-2',),
        ('crystal asymmetry', None),
        ('crystal aspect ratio', None),
        ('emissivity', None),
        ('absorption', None),
        ('backward scattering', None),
        ('side scattering', None),
        ]
    }

ENUMERATIONS['single_scattering_properties_methods'] = {
    'description': 'Methods for calculating single scattering properties of ice crystals',
    'is_open': True,
    'members': [
        ('T-Matrix', 'an exact method'),
        ('geometrical optics', 'for particles that are much larger than the wavelength of light'),
        ('finite difference time domain', 'FDTD'),
        ('anomalous diffraction approximation', 'ADA'),
    ]
}

ENUMERATIONS['cloud_liquid_properties'] = {
    'description': 'Radiative properties of liquid droplets in clouds',
    'is_open': True,
    'members': [
        ('droplet scattering', None),
        ('droplet absorption', None),
        ('cloud droplet number concentration', 'CDNC'),
        ('effective cloud droplet radii', None),
        ('droplet size distribution', None),
        ('liquid water path', 'Integrated liquid water path through the cloud kg m-2',),
        ('broadband reflectivity', 'albedo'),
        ('broadband transmissivity', None),
        ('broadband absorbtivity', None),
        ]
    }

ENUMERATIONS['longwave_scheme_type'] = {
    'description': 'Type of scheme used for longwave radiation parameterisation',
    'is_open': True,
    'members': [
        ('wide-band model', None),
        ('wide-band model (Morcrette)', None),
        ('K-correlated', None),
        ('K-correlated (RRTM)', None),
        ]
    }

ENUMERATIONS['longwave_scheme_method'] = {
    'description': 'Method for the radiative transfer calculations used in the longwave scheme',
    'is_open': True,
    'members': [
        ('two-stream', None),
        ('layer interaction', None),
        ('adaptive', 'exploits spatial and temporal correlations in optical characteristics')
        ]
    }

ENUMERATIONS['shortwave_scheme_type'] = {
    'description': 'Type of scheme used for shortwave radiation parameterisation',
    'is_open': True,
    'members': [
        ('wide-band model', None),
        ('wide-band model (Fouquart)', None),
        ('bulk-scheme', 'highly parameterized methods that use bulk expressions'),
        ('two-stream', None),
        ('two-stream (delta-Eddington)', 'approximation for solar radiation calculations'),
        ]
    }
